#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import traceback
from datetime import datetime
from typing import List, Dict

from fim.slivers.base_sliver import BaseSliver

from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
from fabric_cf.actor.core.apis.abc_actor_identity import ABCActorIdentity
from fabric_cf.actor.core.apis.abc_base_plugin import ABCBasePlugin
from fabric_cf.actor.core.apis.abc_authority_reservation import ABCAuthorityReservation
from fabric_cf.actor.core.apis.abc_broker_reservation import ABCBrokerReservation
from fabric_cf.actor.core.apis.abc_client_callback_proxy import ABCClientCallbackProxy
from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation
from fabric_cf.actor.core.apis.abc_controller_callback_proxy import ABCControllerCallbackProxy
from fabric_cf.actor.core.apis.abc_controller_reservation import ABCControllerReservation
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import SliceNotFoundException
from fabric_cf.actor.core.kernel.failed_rpc import FailedRPC
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.kernel.kernel import Kernel
from fabric_cf.actor.core.common.exceptions import KernelException
from fabric_cf.actor.core.kernel.request_types import RequestTypes
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.kernel.sequence_comparison_codes import SequenceComparisonCodes
from fabric_cf.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric_cf.actor.core.container.maintenance import Site
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.update_data import UpdateData
from fabric_cf.actor.security.access_checker import AccessChecker
from fabric_cf.actor.security.auth_token import AuthToken
from fabric_cf.actor.security.pdp_auth import ActionId


class KernelWrapper:
    """
    KernelWrapper is responsible for validating the arguments to internal kernel methods before
    invoking these calls. The internal kernel methods can only be invoked through
    an instance of the kernel wrapper.
    """
    def __init__(self, *, actor: ABCActorMixin, plugin: ABCBasePlugin, policy: ABCPolicy):
        # The actor linked by the wrapper to the kernel.
        self.actor = actor
        # The kernel instance.
        self.kernel = Kernel(plugin=plugin, policy=policy, logger=actor.get_logger())
        # Logger.
        self.logger = actor.get_logger()

    def await_nothing_pending(self):
        """
        Blocks until there are no more reservations in a pending state.
        @throws InterruptedException in case of error
        """
        self.kernel.await_nothing_pending()

    def claim_delegation_request(self, *, delegation: ABCDelegation, caller: AuthToken,
                                 callback: ABCClientCallbackProxy):
        """
        Processes a request to claim a pre-advertised delegation

        Role: Broker

        @param delegation delegation describing the claim request
        @param caller caller identity
        @param callback callback proxy
        @raises Exception in case of error
        """
        if delegation is None or caller is None or callback is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        # Note: for claim we do not need the slice object, so we use
        # validate_delegation(delegation_id) instead of validate_delegation(delegation).
        exported = self.kernel.validate_delegation(did=delegation.get_delegation_id())
        exported.prepare(callback=callback, logger=self.logger)
        self.kernel.claim_delegation(delegation=exported)

    def reclaim_delegation_request(self, *, delegation: ABCDelegation, caller: AuthToken,
                                   callback: ABCClientCallbackProxy):
        """
        Processes a request to re-claim a pre-claimed delegation

        Role: Broker

        @param delegation delegation describing the reclaim request
        @param caller caller identity
        @param callback callback proxy
        @raises Exception in case of error
        """
        if delegation is None or caller is None or callback is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        # Note: for claim we do not need the slice object, so we use
        # validate(ReservationID) instead of validate(Reservation).
        exported = self.kernel.validate_delegation(did=delegation.get_delegation_id())
        exported.prepare(callback=callback, logger=self.logger)
        self.kernel.reclaim_delegation(delegation=exported)

    def fail(self, *, rid: ID, message: str):
        """
        Fails the specified reservation.
        @param rid reservation id
        @param message message
        @throws Exception in case of error
        """
        if rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.validate(rid=rid)
        self.kernel.fail(reservation=target, message=message)

    def fail_delegation(self, *, did: str, message: str):
        """
        Fails the specified fail_delegation.
        @param did fail_delegation id
        @param message message
        @throws Exception in case of error
        """
        if did is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.validate_delegation(did=did)
        self.kernel.fail_delegation(delegation=target, message=message)

    def close(self, *, rid: ID):
        """
        Closes the reservation, potentially initiating a close request to another
        actor. If the reservation has concrete resources bound to it, this method
        may return before all close operations have completed. Check the
        reservation state to determine when close completes.
        @param rid identifier of reservation to close
        @throws Exception in case of error
        """
        if rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)
        target = self.kernel.validate(rid=rid)
        # NOTE: this call does not require access control check, since
        # it is executed in the context of the actor represented by KernelWrapper.
        self.kernel.close(reservation=target)

    def close_delegation(self, *, did: str):
        """
        Closes the delegation, potentially initiating a close request to another
        actor. If the delegation has concrete resources bound to it, this method
        may return before all close operations have completed. Check the
        delegation state to determine when close completes.
        @param did identifier of reservation to close
        @throws Exception in case of error
        """
        if did is None:
            raise KernelException(Constants.INVALID_ARGUMENT)
        target = self.kernel.validate_delegation(did=did)
        # NOTE: this call does not require access control check, since
        # it is executed in the context of the actor represented by KernelWrapper.
        self.kernel.close_delegation(delegation=target)

    def close_slice_reservations(self, *, slice_id: ID):
        """
        Close Slice reservations
        @param slice_id slice id
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        if not self.kernel.is_known_slice(slice_id=slice_id):
            raise SliceNotFoundException(slice_id=str(slice_id))

        reservations = self.get_reservations(slice_id=slice_id)
        for r in reservations:
            try:
                self.kernel.close(reservation=r)
            except Exception as e:
                self.logger.error("Error during close: {}".format(e))

    def close_request(self, *, reservation: ABCReservationMixin, caller: AuthToken, compare_sequence_numbers: bool):
        """
        Processes an incoming request to close a reservation.
        Role: Authority
        @param reservation reservation to close.
        @param caller caller identity
        @param compare_sequence_numbers if true, the incoming sequence number will
                   be compared to the local sequence number to detect fresh
                   requests, if false, no comparison will be performed.
        @throws Exception in case of error
        """
        if reservation is None or caller is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        if compare_sequence_numbers:
            target = self.kernel.validate(rid=reservation.get_reservation_id())
            sequence_number_compare = self.kernel.compare_and_update_ignore_pending(incoming=reservation,
                                                                                    current=target)
            if sequence_number_compare == SequenceComparisonCodes.SequenceGreater:
                self.kernel.close(reservation=target)
            elif sequence_number_compare == SequenceComparisonCodes.SequenceSmaller:
                self.logger.warning("closeRequest with a smaller sequence number")
            elif sequence_number_compare == SequenceComparisonCodes.SequenceEqual:
                self.logger.warning("duplicate closeRequest")
                self.kernel.handle_duplicate_request(current=target, operation=RequestTypes.RequestClose)
        else:
            target = self.kernel.validate(reservation=reservation)
            self.kernel.close(reservation=target)

    def advertise(self, *, delegation: ABCDelegation, client: AuthToken):
        """
        Initiates a ticket export.
        Role: Broker or Authority
        Prepare/hold a ticket for "will call" claim by a client.
        @param delegation reservation to be exported
        @param client client identity
        @throws Exception in case of error
        """
        if delegation is None or delegation.get_slice_object() is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        delegation.prepare(callback=None, logger=self.logger)
        self.handle_delegate(delegation=delegation, identity=client)

    def extend_lease(self, *, reservation: ABCControllerReservation):
        """
        Initiates a request to extend a lease.
        Role: Controller/Orchestrator
        @param reservation reservation describing the extend request
        @raises Exception in case of error
        """
        if reservation is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.validate(rid=reservation.get_reservation_id())

        if target is None:
            self.logger.error("extendLease for a reservation not registered with the kernel")
            return

        target.validate_redeem()
        self.kernel.extend_lease(reservation=reservation)

    def extend_lease_request(self, *, reservation: ABCAuthorityReservation, caller: AuthToken,
                             compare_sequence_numbers: bool, callback: ABCControllerCallbackProxy = None):
        """
        Processes an incoming request for a lease extension.
        Role: Authority
        @param reservation reservation representing the lease extension request.
        @param caller caller identity
        @param compare_sequence_numbers if true, the incoming sequence number will
                   be compared to the local sequence number to detect fresh
                   requests, if false, no comparison will be performed.
        @param callback callback to be used to send failure response when the request fails.
        @throws Exception in case of error
        """
        if reservation is None or caller is None:
            raise KernelException(Constants.INVALID_ARGUMENT)
        try:
            self.__authorize_request(action_id=ActionId.renew, reservation=reservation,
                                     sliver=reservation.get_requested_resources().get_sliver(),
                                     lease_end_time=reservation.get_requested_term().get_end_time())

            if compare_sequence_numbers:
                reservation.validate_incoming()
                target = self.kernel.validate(rid=reservation.get_reservation_id())

                sequence_compare = self.kernel.compare_and_update(incoming=reservation, current=target)
                if sequence_compare == SequenceComparisonCodes.SequenceGreater:
                    target.prepare_extend_lease()
                    self.kernel.extend_lease(reservation=target)

                elif sequence_compare == SequenceComparisonCodes.SequenceSmaller:
                    self.logger.warning("extendLeaseRequest with a smaller sequence number")

                elif sequence_compare == SequenceComparisonCodes.SequenceEqual:
                    self.logger.warning("duplicate extendLease request")
                    self.kernel.handle_duplicate_request(current=target, operation=RequestTypes.RequestExtendLease)
            else:
                target = self.kernel.validate(rid=reservation.get_reservation_id())
                self.kernel.extend_lease(reservation=target)
        except Exception as e:
            self.logger.error(f"Exception occurred {e}")
            self.logger.error(traceback.format_exc())
            self.fail_notify(reservation=reservation, caller=caller, callback=callback, reason=str(e))

    def modify_lease_request(self, *, reservation: ABCAuthorityReservation, caller: AuthToken,
                             compare_sequence_numbers: bool, callback: ABCControllerCallbackProxy = None,):
        """
        Processes an incoming request for a lease modification.
        Role: Authority
        @param reservation reservation representing the lease modification request.
        @param caller caller identity
        @param compare_sequence_numbers if true, the incoming sequence number will
                   be compared to the local sequence number to detect fresh
                   requests, if false, no comparison will be performed.
        @param callback callback to be used to send failure response when the request fails.
        @throws Exception in case of error
        """
        if reservation is None or caller is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        try:
            self.__authorize_request(action_id=ActionId.modify, reservation=reservation,
                                     sliver=reservation.get_requested_resources().get_sliver(),
                                     lease_end_time=reservation.get_requested_term().get_end_time())

            if compare_sequence_numbers:
                reservation.validate_incoming()
                target = self.kernel.validate(rid=reservation.get_reservation_id())

                sequence_compare = self.kernel.compare_and_update(incoming=reservation, current=target)

                if sequence_compare == SequenceComparisonCodes.SequenceGreater:
                    target.prepare_modify_lease()
                    self.kernel.modify_lease(reservation=target)

                elif sequence_compare == SequenceComparisonCodes.SequenceSmaller:
                    self.logger.warning("modifyLeaseRequest with a smaller sequence number")

                elif sequence_compare == SequenceComparisonCodes.SequenceEqual:
                    self.logger.warning("duplicate extendLease request")
                    self.kernel.handle_duplicate_request(current=target, operation=RequestTypes.RequestModifyLease)
            else:
                target = self.kernel.validate(rid=reservation.get_reservation_id())
                self.kernel.modify_lease(reservation=target)
        except Exception as e:
            self.logger.error(f"Exception occurred {e}")
            self.logger.error(traceback.format_exc())
            self.fail_notify(reservation=reservation, caller=caller, callback=callback, reason=str(e))

    def extend_reservation(self, *, rid: ID, resources: ResourceSet, term: Term,
                           dependencies: List[ABCReservationMixin] = None) -> int:
        """
        Extends the reservation with the given resources and term.
        @param rid identifier of reservation to extend
        @param resources resources for extension
        @param term term for extension
        @param dependencies: dependencies
        @return 0 on success, a negative exit code on error
        @throws Exception in case of error
        """
        if rid is None or resources is None or term is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        return self.kernel.extend_reservation(rid=rid, resources=resources, term=term, dependencies=dependencies)

    def extend_ticket(self, *, reservation: ABCClientReservation):
        """
        Initiates a request to extend a ticket.
        Role: Broker or Controller
        @param reservation reservation describing the ticket extension request
        @throws Exception in case of error
        """
        if reservation is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.validate(rid=reservation.get_reservation_id())

        if target is None:
            raise KernelException("extendTicket on a reservation not registered with the kernel")

        target.validate_outgoing()
        self.kernel.extend_ticket(reservation=target)

    def extend_ticket_request(self, *, reservation: ABCBrokerReservation, caller: AuthToken,
                              compare_sequence_numbers: bool, callback: ABCClientCallbackProxy = None):
        """
        Processes an incoming request for a ticket extension.
        Role: Broker
        @param reservation reservation representing the ticket extension request.
        @param caller caller identity
        @param compare_sequence_numbers if true, the incoming sequence number will
                   be compared to the local sequence number to detect fresh
                   requests, if false, no comparison will be performed.
        @param callback callback to be used to send the response back if the request failed
        @throws Exception in case of error
        """
        self.logger.debug("extend_ticket_request")
        if reservation is None or caller is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        try:
            self.__authorize_request(action_id=ActionId.renew, reservation=reservation,
                                     sliver=reservation.get_requested_resources().get_sliver(),
                                     lease_end_time=reservation.get_requested_term().get_end_time())

            if compare_sequence_numbers:
                target = self.kernel.validate(rid=reservation.get_reservation_id())

                sequence_compare = self.kernel.compare_and_update(incoming=reservation, current=target)

                if sequence_compare == SequenceComparisonCodes.SequenceGreater:
                    self.logger.debug("extend_ticket SequenceGreater")
                    self.kernel.extend_ticket(reservation=target)

                elif sequence_compare == SequenceComparisonCodes.SequenceInProgress:
                    self.logger.warning("New request for a reservation with a pending action")

                elif sequence_compare == SequenceComparisonCodes.SequenceSmaller:
                    self.logger.warning("Incoming extendTicket request has smaller sequence number")

                elif sequence_compare == SequenceComparisonCodes.SequenceEqual:
                    self.logger.warning("Duplicate extendTicket request")
                    self.kernel.handle_duplicate_request(current=target, operation=RequestTypes.RequestExtendTicket)
            else:
                target = self.kernel.validate(rid=reservation.get_reservation_id())
                self.logger.debug("extend_ticket No sequence number comparison")
                self.extend_ticket(reservation=target)
        except Exception as e:
            self.logger.error(f"Exception occurred {e}")
            self.logger.error(traceback.format_exc())
            self.fail_notify(reservation=reservation, caller=caller, callback=callback, reason=str(e))

    def modify_lease(self, *, reservation: ABCControllerReservation):
        """
        Initiates a request to modify a lease.
        Role: Controller
        @param reservation reservation describing the modify request
        @throws Exception in case of error
        """
        if reservation is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.validate(rid=reservation.get_reservation_id())

        if target is None:
            self.logger.error("modifyLease for a reservation not registered with the kernel")

        target.validate_redeem()
        self.kernel.modify_lease(reservation=target)

    def relinquish_request(self, *, reservation: ABCBrokerReservation, caller: AuthToken):
        if reservation is None or caller is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        reservation.validate_incoming()

        target = self.kernel.soft_validate(rid=reservation.get_reservation_id())
        if target is None:
            self.logger.info("Relinquish for non-existent reservation. Reservation has already been closed. "
                             "Nothing to relinquish")
            return

        sequence_compare = self.kernel.compare_and_update(incoming=reservation, current=target)

        if sequence_compare == SequenceComparisonCodes.SequenceGreater or \
                sequence_compare == SequenceComparisonCodes.SequenceInProgress:
            self.kernel.close(reservation=target)

        elif sequence_compare == SequenceComparisonCodes.SequenceSmaller:
            self.logger.warning("Incoming relinquish request has smaller sequence number")

        elif sequence_compare == SequenceComparisonCodes.SequenceEqual:
            self.logger.warning("Duplicate relinquish request")
            self.kernel.handle_duplicate_request(current=target, operation=RequestTypes.RequestRelinquish)

    def get_client_slices(self) -> List[ABCSlice]:
        """
        Returns all client slices registered with the kernel.
        @return an array of client slices registered with the kernel
        """
        return self.kernel.get_client_slices()

    def get_inventory_slices(self) -> List[ABCSlice]:
        """
        Returns all inventory slices registered with the kernel.
        @return an array of inventory slices registered with the kernel
        """
        return self.kernel.get_inventory_slices()

    def get_delegation(self, *, did: str) -> ABCDelegation:
        """
        Returns the delegation with the given delegation identifier.
        @param did delegation identifier
        @return delegation with the given delegation identifier
        """
        if did is None:
            raise KernelException(Constants.INVALID_ARGUMENT)
        return self.kernel.get_delegation(did=did)

    def get_reservation(self, *, rid: ID) -> ABCReservationMixin:
        """
        Returns the reservation with the given reservation identifier.
        @param rid reservation identifier
        @return reservation with the given reservation identifier
        """
        if rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)
        return self.kernel.get_reservation(rid=rid)

    def get_reservations(self, *, slice_id: ID) -> List[ABCReservationMixin]:
        """
        Returns all reservations in the given slice
        @param slice_id identifier of slice
        @return an array of all reservations in the slice
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        return self.kernel.get_reservations(slice_id=slice_id)

    def get_slice(self, *, slice_id: ID) -> ABCSlice:
        """
        Returns the slice with the given name.
        @param slice_id identifier of slice to return
        @return the requested slice or null if no slice with the requested name
                is registered with the kernel.
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)
        return self.kernel.get_slice(slice_id=slice_id)

    def get_slices(self) -> list:
        """
        Returns all slice registered with the kernel.
        @return an array of slices registered with the kernel
        """
        return self.kernel.get_slices()

    def handle_delegate(self, *, delegation: ABCDelegation, identity: AuthToken):
        """
        Handles a delegation. Called from both AM and Broker.

        @param delegation the delegation
        @param identity caller identity

        @throws Exception in case of error
        """
        if delegation.get_slice_object() is None or delegation.get_slice_object().get_name() is None or \
                delegation.get_slice_object().get_slice_id() is None or delegation.get_delegation_id() is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        # Obtain the previously created slice or create a new slice. When this
        # function returns we will have a slice object that is registered with the kernel
        s = self.kernel.get_slice(slice_id=delegation.get_slice_id())
        if s is None:
            self.kernel.register_slice(slice_object=delegation.get_slice_object())

        # Determine if this is a new or an already existing reservation. We
        # will register new reservations and call reserve for them. For
        # existing reservations we will perform amendReserve. XXX: Note, if the
        # caller makes two concurrent requests for a new reservation, it is
        # possible that kernel.softValidate can return null for both, yet only
        # one of them will succeed to register itself. For the second the
        # kernel will throw an exception. This is a limitation that does not
        # seem to be too much of a problem and can be resolved using the same
        # technique we use for creating/registering slices.
        temp = self.kernel.soft_validate_delegation(delegation=delegation)

        if temp is None:
            self.kernel.register_delegation(delegation=delegation)
            self.kernel.delegate(delegation=delegation)
        else:
            self.kernel.amend_delegate(delegation=temp)

    def handle_reserve(self, *, reservation: ABCReservationMixin, identity: AuthToken, create_new_slice: bool):
        """
        Handles a reserve, i.e., obtain a new ticket or lease. Called from both
        client and server side code. If the slice does not exist it will create
        and register a slice (if create_new_slice is true (server
        side)). If the slice does not exist and create_new_slice is
        false (client side), it will register the slice contained in the
        reservation object.
        @param reservation the reservation
        @param identity caller identity
        @param create_new_slice true -> creates a new slice object, false -> reuses
                   the slice object in the reservation. This flag is considered
                   only if the slice referenced by the reservation is not
                   registered with the kernel.
        @throws Exception in case of error
        """
        if reservation.get_slice() is None or reservation.get_slice().get_name() is None or \
                reservation.get_slice().get_slice_id() is None or reservation.get_reservation_id() is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        # Obtain the previously created slice or create a new slice. When this
        # function returns we will have a slice object that is registered with the kernel
        self.kernel.get_or_create_local_slice(identity=identity, reservation=reservation,
                                              create_new_slice=create_new_slice)

        # Determine if this is a new or an already existing reservation. We
        # will register new reservations and call reserve for them. For
        # existing reservations we will perform amendReserve. XXX: Note, if the
        # caller makes two concurrent requests for a new reservation, it is
        # possible that kernel.softValidate can return null for both, yet only
        # one of them will succeed to register itself. For the second the
        # kernel will throw an exception. This is a limitation that does not
        # seem to be too much of a problem and can be resolved using the same
        # technique we use for creating/registering slices.
        temp = self.kernel.soft_validate(reservation=reservation)

        if temp is None:
            self.kernel.register_reservation(reservation=reservation)
            self.kernel.reserve(reservation=reservation)
        else:
            self.kernel.amend_reserve(reservation=temp)

    def handle_update_reservation(self, *, reservation: ABCReservationMixin, auth: AuthToken):
        """
        Amend a reservation request or initiation, i.e., to issue a new bid on a
        previously filed request.
        @param reservation the reservation
        @param auth the slice owner
        @throws Exception in case of error
        """
        self.kernel.amend_reserve(reservation=reservation)

    def query(self, *, properties: dict, caller: AuthToken):
        """
        Processes an incoming query request.
        @param properties query
        @param caller caller identity
        @return query response
        """

        return self.kernel.query(properties=properties)

    def redeem(self, *, reservation: ABCControllerReservation):
        """
        Initiates a request to redeem a ticketed reservation.
        Role: Controller
        @param reservation the reservation being redeemed
        @throws Exception in case of error
        """
        if reservation is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        slice_object = reservation.get_slice()

        for r in slice_object.get_reservations().values():
            self.logger.debug("redeem() Reservation {} is in state: {}".format(r.get_reservation_id(),
                                                                               ReservationStates(r.get_state()).name))

        target = self.kernel.validate(rid=reservation.get_reservation_id())

        if target is None:
            self.logger.error("Redeem on a reservation not registered with the kernel")

        target.validate_redeem()
        self.kernel.redeem(reservation=target)

    def redeem_request(self, *, reservation: ABCAuthorityReservation, caller: AuthToken,
                       callback: ABCControllerCallbackProxy, compare_sequence_numbers: bool):
        """
        Processes an incoming request for a new lease.
        Role: Authority
        @param reservation reservation representing the lease request. Must
                   contain a valid ticket.
        @param caller caller identity
        @param callback callback object
        @param compare_sequence_numbers if true, the incoming sequence number will
                   be compared to the local sequence number to detect fresh
                   requests, if false, no comparison will be performed.
        @throws Exception in case of error
        """
        if reservation is None or caller is None or callback is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        try:
            self.__authorize_request(action_id=ActionId.redeem, reservation=reservation,
                                     sliver=reservation.get_requested_resources().get_sliver(), lease_end_time=None)

            if compare_sequence_numbers:
                reservation.validate_incoming()
                reservation.get_slice().set_client()

                s = self.kernel.get_slice(slice_id=reservation.get_slice().get_slice_id())

                if s is not None:
                    target = self.kernel.soft_validate(rid=reservation.get_reservation_id())

                    if target is not None:
                        self.kernel.handle_duplicate_request(current=target, operation=RequestTypes.RequestRedeem)

                reservation.prepare(callback=callback, logger=self.logger)
                self.handle_reserve(reservation=reservation, identity=caller, create_new_slice=True)
            else:
                reservation.set_logger(logger=self.logger)
                self.handle_reserve(reservation=reservation, identity=reservation.get_client_auth_token(),
                                    create_new_slice=True)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Exception occurred in processing redeem request {}".format(e))
            self.fail_notify(reservation=reservation, caller=caller, callback=callback, reason=str(e))

    def register_reservation(self, *, reservation: ABCReservationMixin):
        """
        Registers the given reservation with the kernel. Adds a database record
        for the reservation. The containing slice should have been previously
        registered with the kernel and no database record should exist. When
        registering a reservation with an existing database record use
        #reregisterReservation(IReservation)}. Only reservations
        that are not closed or failed can be registered. Closed or failed
        reservations will be ignored.
        @param reservation the reservation to register
        @throws IllegalArgumentException when the passed in argument is illegal
        @throws Exception if the reservation has already been registered with the
                    kernel.
        @throws RuntimeException when a database error occurs. In this case the
                    reservation will be unregistered from the kernel data
                    structures.
        """
        if reservation is None or not isinstance(reservation, ABCReservationMixin):
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.register_reservation(reservation=reservation)

    def register_delegation(self, *, delegation: ABCDelegation):
        """
        Registers the given delegation with the kernel. Adds a database record
        for the delegation. The containing slice should have been previously
        registered with the kernel and no database record should exist. When
        registering a delegation with an existing database record use
        #re_register_delegation(ABCDelegation)}. Only delegations
        that are not closed or failed can be registered. Closed or failed
        delegations will be ignored.
        @param delegation the delegation to register
        @throws IllegalArgumentException when the passed in argument is illegal
        @throws Exception if the delegation has already been registered with the
                    kernel.
        @throws RuntimeException when a database error occurs. In this case the
                    delegation will be unregistered from the kernel data
                    structures.
        """
        if delegation is None or not isinstance(delegation, ABCDelegation):
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.register_delegation(delegation=delegation)

    def modify_slice(self, *, slice_object: ABCSlice):
        """
        Modify the slice registered with the kernel
        @param slice_object slice_object
        @throws Exception in case of error
        """
        if slice_object is None or slice_object.get_slice_id() is None or not isinstance(slice_object, ABCSlice):
            raise KernelException("Invalid argument {}".format(slice_object))

        self.kernel.modify_slice(slice_object=slice_object)

    def register_slice(self, *, slice_object: ABCSlice):
        """
        Registers the slice with the kernel: adds the slice object to the kernel
        data structures and adds a database record for the slice.
        @param slice_object slice to register
        @throws Exception if the slice is already registered or a database error
                    occurs. If a database error occurs, the slice will be
                    unregistered.
        """
        if slice_object is None or slice_object.get_slice_id() is None or not isinstance(slice_object, ABCSlice):
            raise KernelException("Invalid argument {}".format(slice_object))

        if slice_object.get_owner() is None:
            slice_object.set_owner(owner=self.actor.get_identity())
        else:
            slice_object.get_owner().name = self.actor.get_name()
            slice_object.get_owner().guid = self.actor.get_guid()

        self.kernel.register_slice(slice_object=slice_object)

    def remove_delegation(self, *, did: str):
        """
        Unregisters the delegation from the kernel data structures and removes
        its record from the database.
        Note:Only failed, closed, or close waiting reservations can be
        removed.
        @param delegation identifier of delegation to remove
        @throws Exception in case of error
        """
        if did is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.remove_delegation(did=did)

    def remove_reservation(self, *, rid: ID):
        """
        Unregisters the reservation from the kernel data structures and removes
        its record from the database.
        Note:Only failed, closed, or close waiting reservations can be
        removed.
        @param rid identifier of reservation to remove
        @throws Exception in case of error
        """
        if rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.remove_reservation(rid=rid)

    def remove_slice(self, *, slice_id: ID):
        """
        Unregisters the slice (if it is registered with the kernel) and removes
        it from the database.

        Note: A slice can be removed only if it contains only closed or
        failed reservations.

        @param slice_id identifier of slice to remove
        @throws Exception in case of error
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.remove_slice(slice_id=slice_id)

    def modify_accept(self, *, slice_id: ID):
        """
        Accept the last modify on the slice

        @param slice_id identifier of slice
        @throws Exception in case of error
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.modify_accept(slice_id=slice_id)

    def re_register_reservation(self, *, reservation: ABCReservationMixin):
        """
         Registers a previously unregistered reservation with the kernel. The
        containing slice should have been previously registered with the kernel
        and a database record for the reservation should exist. When registering
        a reservation without and existing database record use
        register_reservation(IReservation). Only reservations
        that are not closed or failed can be registered. Closed or failed
        reservations will be ignored.
        @param reservation the reservation to reregister
        @throws IllegalArgumentException when the passed in argument is illegal
        @throws Exception if the reservation has already been registered with the
                    kernel or the reservation does not have a database record. In
                    the latter case the reservation will be unregistered from the
                    kernel data structures.
        @throws RuntimeException if a database error occurs
        """
        if reservation is None or not isinstance(reservation, ABCReservationMixin):
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.re_register_reservation(reservation=reservation)

    def re_register_delegation(self, *, delegation: ABCDelegation):
        """
         Registers a previously unregistered delegation with the kernel.
        @param delegation the delegation to reregister
        @throws IllegalArgumentException when the passed in argument is illegal
        @throws Exception if the delegation has already been registered with the
                    kernel or the delegation does not have a database record. In
                    the latter case the delegation will be unregistered from the
                    kernel data structures.
        @throws RuntimeException if a database error occurs
        """
        if delegation is None or not isinstance(delegation, ABCDelegation):
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.re_register_delegation(delegation=delegation)

    def re_register_slice(self, *, slice_object: ABCSlice):
        """
        Registers the slice with the kernel: adds the slice object to the kernel
        data structures. The slice object must have an existing database record.
        @param slice_object slice to register
        @throws Exception if the slice is already registered or a database error
                    occurs. If a database error occurs, the slice will be
                    unregistered.
        """
        if slice_object is None or slice_object.get_slice_id() is None or not isinstance(slice_object, ABCSlice):
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.re_register_slice(slice_object=slice_object)

    def tick(self):
        """
        Checks all reservations for completions or problems. We might want to do
        these a few at a time.
        @throws Exception in case of error
        """
        try:
            self.kernel.tick()
        except Exception as e:
            self.logger.error("Tick error: {}".format(e))
            self.logger.error(traceback.format_exc())

    def delegate(self, *, delegation: ABCDelegation, destination: ABCActorIdentity):
        """
        Initiates a delegate request. If the exported flag is set, this is a claim
        on a pre-reserved "will call" ticket.
        Role: Broker or Controller.
        @param delegation delegation parameters for ticket request
        @param destination identity of the actor the request must be sent to
        @throws Exception in case of error
        """
        if delegation is None or destination is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        callback = ActorRegistrySingleton.get().get_callback(protocol=Constants.PROTOCOL_KAFKA,
                                                             actor_name=destination.get_name())
        if callback is None:
            raise KernelException("Unsupported")

        delegation.prepare(callback=callback, logger=self.logger)
        delegation.validate_outgoing()
        self.handle_delegate(delegation=delegation, identity=destination.get_identity())

    def ticket(self, *, reservation: ABCClientReservation, destination: ABCActorIdentity):
        """
        Initiates a ticket request. If the exported flag is set, this is a claim
        on a pre-reserved "will call" ticket.
        Role: Broker or Controller.
        @param reservation reservation parameters for ticket request
        @param destination identity of the actor the request must be sent to
        @throws Exception in case of error
        """
        if reservation is None or destination is None or not isinstance(reservation, ABCClientReservation):
            raise KernelException(Constants.INVALID_ARGUMENT)

        protocol = reservation.get_broker().get_type()
        callback = ActorRegistrySingleton.get().get_callback(protocol=protocol, actor_name=destination.get_name())
        if callback is None:
            raise KernelException("Unsupported")

        reservation.prepare(callback=callback, logger=self.logger)
        reservation.validate_outgoing()
        self.handle_reserve(reservation=reservation, identity=destination.get_identity(), create_new_slice=False)

    def ticket_request(self, *, reservation: ABCBrokerReservation, caller: AuthToken, callback: ABCClientCallbackProxy,
                       compare_seq_numbers: bool):
        """
        Processes an incoming request for a new ticket.
        Role: Broker
        @param reservation reservation representing the ticket request
        @param caller caller identity
        @param callback callback object
        @param compare_seq_numbers if true, the incoming sequence number will
                   be compared to the local sequence number to detect fresh
                   requests, if false, no comparison will be performed.
        @throws Exception in case of error
        """
        if reservation is None or caller is None or callback is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        try:
            self.__authorize_request(action_id=ActionId.ticket, reservation=reservation,
                                     sliver=reservation.get_requested_resources().get_sliver(), lease_end_time=None)

            if compare_seq_numbers:
                reservation.validate_incoming()
                # Mark the slice as client slice.
                reservation.get_slice().set_client()

                # This reservation has just arrived at this actor from another
                # actor. Attach the current actor object to the reservation so
                # that operations on this reservation can get access to it.
                reservation.set_actor(actor=self.kernel.get_plugin().get_actor())

                # If the slice referenced in the reservation exists, then we
                # must check the incoming sequence number to determine if this
                # request is fresh.
                s = self.kernel.get_slice(slice_id=reservation.get_slice().get_slice_id())

                if s is not None:
                    target = self.kernel.soft_validate(rid=reservation.get_reservation_id())
                    if target is not None:
                        seq_num = self.kernel.compare_and_update(incoming=reservation, current=target)
                        if seq_num == SequenceComparisonCodes.SequenceGreater:
                            self.handle_update_reservation(reservation=target, auth=caller)

                        elif seq_num == SequenceComparisonCodes.SequenceSmaller:
                            self.logger.warning("Incoming request has a smaller sequence number")

                        elif seq_num == SequenceComparisonCodes.SequenceInProgress:
                            self.logger.warning("New request for a reservation with a pending action")

                        elif seq_num == SequenceComparisonCodes.SequenceEqual:
                            self.kernel.handle_duplicate_request(current=target, operation=RequestTypes.RequestTicket)
                    else:
                        # This is a new reservation. No need to check sequence numbers
                        reservation.prepare(callback=callback, logger=self.logger)
                        self.handle_reserve(reservation=reservation, identity=caller, create_new_slice=True)
                else:
                    # New reservation for a new slice.
                    reservation.prepare(callback=callback, logger=self.logger)
                    self.handle_reserve(reservation=reservation, identity=caller, create_new_slice=True)
            else:
                # This is most likely a reservation being recovered. Do not compare sequence numbers,
                # just trigger reserve.
                reservation.set_logger(logger=self.logger)
                self.handle_reserve(reservation=reservation, identity=reservation.get_client_auth_token(),
                                    create_new_slice=True)
        except Exception as e:
            self.logger.error(f"Exception occurred {e}")
            self.logger.error(traceback.format_exc())
            self.fail_notify(reservation=reservation, caller=caller, callback=callback, reason=str(e))

    def unregister_reservation(self, *, rid: ID):
        """
        Unregisters the reservation from the kernel data structures.
         Note:does not remove the reservation database record.
         Note:Only failed, closed, or close waiting reservations can be
        unregistered.
        @param rid identifier for reservation to unregister
        @throws Exception in case of error
        """
        if rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.unregister_reservation(rid=rid)

    def unregister_slice(self, *, slice_id: ID):
        """
        Unregisters the slice and releases any resources that it may hold.
        Note: A slice can be unregistered only if it contains only closed
        or failed reservations.
        @param slice_id identifier of slice to unregister
        @throws Exception in case of error
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        self.kernel.unregister_slice(slice_id=slice_id)

    def update_lease(self, *, reservation: ABCReservationMixin, update_data: UpdateData, caller: AuthToken):
        """
        Handles a lease update from an authority.
        Role: Controller
        @param reservation reservation describing the update
        @param update_data status of the update
        @param caller identity of the caller
        @throws Exception in case of error
        """
        if reservation is None or update_data is None or caller is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.validate(rid=reservation.get_reservation_id())
        reservation.validate_incoming()
        self.kernel.update_lease(reservation=target, update=reservation, update_data=update_data)

    def update_ticket(self, *, reservation: ABCReservationMixin, update_data: UpdateData, caller: AuthToken):
        """
        Handles a ticket update from upstream broker.
        Role: Agent or Controller.
        @param reservation reservation describing the update
        @param update_data status of the update
        @param caller identity of the caller
        @throws Exception in case of error
        """
        if reservation is None or update_data is None or caller is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.validate(rid=reservation.get_reservation_id())
        reservation.validate_incoming_ticket()
        self.kernel.update_ticket(reservation=target, update=reservation, update_data=update_data)

    def update_delegation(self, *, delegation: ABCDelegation, update_data: UpdateData, caller: AuthToken):
        """
        Handles a delegation update from upstream broker.
        Role: Agent or Controller.
        @param delegation delegation describing the update
        @param update_data status of the update
        @param caller identity of the caller
        @throws Exception in case of error
        """
        if delegation is None or update_data is None or caller is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.validate_delegation(did=delegation.get_delegation_id())
        delegation.validate_incoming()
        self.kernel.update_delegation(delegation=target, update=delegation, update_data=update_data)

    def process_failed_rpc(self, *, rid: ID, rpc: FailedRPC):
        if rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        target = self.kernel.soft_validate(rid=rid)
        if target is None:
            self.logger.warning("Could not find reservation #{} while processing a failed RPC.".format(rid))
            return
        self.kernel.handle_failed_rpc(reservation=target, rpc=rpc)

    def fail_notify(self, *, reservation: ABCReservationMixin, reason: str, caller: AuthToken,
                    callback: ABCControllerCallbackProxy):
        """
        Handles a fail notification from an authority or broker.
        Broker:
            - ExtendTicket or Ticket
        Authority:
            - ExtendLease or ModifyLease or Redeem
        """
        try:
            reservation.get_slice().set_owner(owner=caller)
            reservation.prepare(callback=callback, logger=self.logger)
            reservation.fail_notify(message=reason)
        except Exception as e:
            self.logger.exception(f"Failed to notify reservation #{reservation.get_reservation_id()} of failure.")

    def __authorize_request(self, *, reservation: ABCReservationMixin, action_id: ActionId, sliver: BaseSliver,
                            lease_end_time: datetime = None):
        slice_object = reservation.get_slice()
        config_props = slice_object.get_config_properties()
        project = config_props.get(Constants.PROJECT_ID, None)
        tags = config_props.get(Constants.TAGS, None)
        email = config_props.get(Constants.CLAIMS_EMAIL, None)

        if project is not None and tags is not None and email is not None:
            tag_list = tags.split(",")
            AccessChecker.check_pdp_access(action_id=action_id, email=email, project=project, tags=tag_list,
                                           resource=sliver, lease_end_time=lease_end_time, logger=self.logger)

    def update_maintenance_mode(self, *, properties: Dict[str, str], sites: List[Site] = None):
        self.kernel.update_maintenance_mode(properties=properties, sites=sites)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import queue
import threading
import time
import traceback

from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.iterable_queue import IterableQueue
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_slice_wrapper import OrchestratorSliceWrapper
from fabric_cf.orchestrator.core.reservation_status_update import ReservationStatusUpdate


class SliceDeferThread:
    """
    This runs as a standalone thread started by Orchestrator and deals with issuing demand for the slivers for
    the newly created slices. The purpose of this thread is to help orchestrator respond back to the create
    without waiting for the slivers to be demanded
    """

    def __init__(self, *, kernel):
        self.slice_queue = queue.Queue()
        self.slice_avail_condition = threading.Condition()
        self.thread_lock = threading.Lock()
        self.thread = None
        self.stopped = False
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.mgmt_actor = kernel.get_management_actor()
        self.kernel = kernel
        self.sut = kernel.get_sut()

    def get_sut(self):
        if self.sut is None:
            self.sut = self.kernel.get_sut()
        return self.sut

    def queue_slice(self, *, controller_slice: OrchestratorSliceWrapper):
        """
        Queue a slice
        :param controller_slice:
        :return:
        """
        try:
            self.slice_queue.put_nowait(controller_slice)
            self.logger.debug(f"Added slice to slices queue {controller_slice.get_slice_id()}")
        except Exception as e:
            self.logger.error(f"Failed to queue slice: {controller_slice.get_slice_id()} e: {e}")

    def start(self):
        """
        Start thread
        :return:
        """
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise OrchestratorException("This SliceDeferThread has already been started")

            self.thread = threading.Thread(target=self.run)
            self.thread.setName(self.__class__.__name__)
            self.thread.setDaemon(True)
            self.thread.start()

        finally:
            self.thread_lock.release()

    def stop(self):
        """
        Stop thread
        :return:
        """
        self.stopped = True
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the SliceDeferThread is running. Interrupting it")
                try:
                    # TODO find equivalent of interrupt
                    with self.slice_avail_condition:
                        self.slice_avail_condition.notify_all()
                    temp.join()
                except Exception as e:
                    self.logger.error(f"Could not join SliceDeferThread thread {e}")
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def run(self):
        """
        Thread main loop
        :return:
        """
        self.logger.debug(f"SliceDeferThread started")
        while True:
            slices = []
            if not self.stopped and self.slice_queue.empty():
                time.sleep(0.001)

            if self.stopped:
                self.logger.info("SliceDeferThread exiting")
                return

            if not self.slice_queue.empty():
                try:
                    for s in IterableQueue(source_queue=self.slice_queue):
                        slices.append(s)
                except Exception as e:
                    self.logger.error(f"Error while adding slice to slice queue! e: {e}")
                    self.logger.error(traceback.format_exc())

            if len(slices) > 0:
                self.logger.debug(f"Processing {len(slices)} slices")
                for s in slices:
                    try:
                        # Process the Slice i.e. Demand the computed reservations i.e. Add them to the policy
                        # Once added to the policy; Actor Tick Handler will do following asynchronously:
                        # 1. Ticket message exchange with broker and
                        # 2. Redeem message exchange with AM once ticket is granted by Broker
                        self.process_slice(controller_slice=s)
                    except Exception as e:
                        self.logger.error(f"Error while processing slice {type(s)}, {e}")
                        self.logger.error(traceback.format_exc())

    def process_slice(self, *, controller_slice: OrchestratorSliceWrapper):
        """
        Demand slice reservations.
        :param controller_slice:
        """
        computed_reservations = controller_slice.get_computed_reservations()

        try:
            controller_slice.lock()
            for reservation in computed_reservations:
                self.logger.debug(f"Issuing demand for reservation: {reservation.get_reservation_id()}")

                if reservation.get_state() != ReservationStates.Unknown.value:
                    self.logger.debug(f"Reservation not in {reservation.get_state()} state, ignoring it")
                    continue

                if not self.mgmt_actor.demand_reservation(reservation=reservation):
                    raise OrchestratorException(f"Could not demand resources: {self.mgmt_actor.get_last_error()}")
                self.logger.debug(f"Reservation #{reservation.get_reservation_id()} demanded successfully")
            '''
            for r in controller_slice.computed_l3_reservations:
                res_status_update = ReservationStatusUpdate(logger=self.logger)
                self.get_sut().add_active_status_watch(watch=ID(uid=r.get_reservation_id()),
                                                       callback=res_status_update)
            '''

            for r in controller_slice.computed_remove_reservations:
                self.logger.debug(f"Issuing close for reservation: {r}")
                self.mgmt_actor.close_reservation(rid=ID(uid=r))

            for rid, modified_res in controller_slice.computed_modify_reservations.items():
                self.logger.debug(f"Issuing extend for modified reservation: {rid}")
                if not self.mgmt_actor.extend_reservation(reservation=ID(uid=rid), sliver=modified_res.sliver,
                                                          new_end_time=None, dependencies=modified_res.dependencies):
                    self.logger.error(f"Could not demand resources: {self.mgmt_actor.get_last_error()}")
                    continue
                self.logger.debug(f"Issued extend for reservation #{rid} successfully")

            for r in controller_slice.computed_modify_properties_reservations:
                self.logger.debug(f"Issuing modify for reservation: {r}")
                self.mgmt_actor.modify_reservation(rid=ID(uid=str(r.get_reservation_id())),
                                                   modified_sliver=r.sliver)

        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Unable to get orchestrator or demand reservation: {}".format(e))
        finally:
            controller_slice.unlock()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING

from fabric_cf.actor.boot.configuration import ActorConfig
from fabric_cf.actor.core.common.exceptions import BrokerException, ExceptionErrorCode
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric_cf.actor.core.time.calendar.broker_calendar import BrokerCalendar
from fabric_cf.actor.core.util.reservation_set import ReservationSet
from fabric_cf.actor.core.apis.abc_broker_reservation import ABCBrokerReservation
from fabric_cf.actor.core.core.broker_policy import BrokerPolicy
from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_broker_mixin import ABCBrokerMixin
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin


class BrokerCalendarPolicy(BrokerPolicy):
    """
    BrokerCalendarPolicy specifies and implements some of the
    broker's base resource allocation and upstream bidding policy.
    """
    def __init__(self, *, actor: ABCBrokerMixin):
        super().__init__(actor=actor)
        # The broker calendar: list of client requests, source reservations, and allocated reservations.
        self.calendar = None
        # Indicates if this actor is initialized
        self.initialized = False

    def add_to_calendar(self, *, reservation: ABCBrokerReservation):
        """
        Records the reservation in the calendar.

        @param reservation reservation
        """
        if reservation.get_approved_resources() is not None and not reservation.is_failed():
            self.calendar.add_outlay(source=reservation.get_source(), client=reservation,
                                     start=reservation.get_approved_term().get_new_start_time(),
                                     end=reservation.get_approved_term().get_end_time())

            if reservation.get_term() is not None:
                self.calendar.remove_closing(reservation=reservation)

            self.calendar.add_closing(reservation=reservation,
                                      cycle=self.clock.cycle(when=reservation.get_approved_term().get_end_time()))

            self.logger.debug("AgentAllocated: units= {} res= {} term= {} "
                              .format(reservation.get_approved_resources().get_units(),
                                      reservation, reservation.get_approved_term()))
        else:
            reservation.fail(message="Either there are no resources on the source or the reservation failed",
                             exception=None)

    def check_pending(self):
        """
        Checks pending bids, and installs successfully completed
        requests in the holdings calendar. Note that the policy module must add
        bids to the pending set, or they may not install in the calendar.

        @throws Exception in case of error
        """
        rvset = self.calendar.get_pending()

        if rvset is None:
            return

        for reservation in rvset.values():
            if not reservation.is_nascent() and reservation.is_no_pending():
                self.logger.debug("Pending request completed {}".format(reservation))

            if not reservation.is_terminal() and reservation.is_renewable():
                cycle = self.get_renew(reservation=reservation)
                reservation.set_renew_time(time=cycle)
                reservation.set_dirty()

            self.calendar.remove_pending(reservation)

    def close(self, *, reservation: ABCReservationMixin):
        if isinstance(reservation, ABCClientReservation):
            rset = self.calendar.get_outlays(source=reservation)
            self.logger.debug("Client reservation; get outlays: {}".format(rset))
            self.actor.close_reservations(reservations=rset)
        else:
            self.logger.debug("Removing reservation from scheduled or in progress list")
            self.calendar.remove_scheduled_or_in_progress(reservation=reservation)

    def closed(self, *, reservation: ABCReservationMixin):
        self.release(reservation=reservation)

    def finish(self, *, cycle: int):
        self.calendar.tick(cycle=cycle)

    def get_closing(self, *, cycle: int) -> ReservationSet:
        return self.calendar.get_closing(cycle=cycle)

    def get_renew(self, *, reservation: ABCClientReservation) -> int:
        """
        Returns the cycle when the reservation must be renewed.

        @param reservation reservation for which to calculate renew time

        @return renew cycle

        @throws Exception in case of error
        """
        raise BrokerException(error_code=ExceptionErrorCode.NOT_IMPLEMENTED)

    def initialize(self, *, config: ActorConfig):
        if not self.initialized:
            super().initialize(config=config)
            self.calendar = BrokerCalendar(clock=self.clock)
            self.initialized = True

    def release(self, *, reservation):
        if isinstance(reservation, ABCBrokerReservation):
            self.logger.debug("Broker reservation")
            source = reservation.get_source()
            if source is not None:
                self.logger.debug("Broker reservation; removing outlay")
                self.calendar.remove_outlay(source=source, client=reservation)

        elif isinstance(reservation, ABCClientReservation):
            self.logger.debug("Client reservation; removing source calendar")
            self.calendar.remove_source_calendar(source=reservation)

    def remove(self, *, reservation: ABCReservationMixin):
        self.calendar.remove(reservation=reservation)

    def revisit(self, *, reservation: ABCReservationMixin):
        super().revisit(reservation=reservation)

        if isinstance(reservation, ABCClientReservation):
            self.revisit_client(reservation=reservation)
        elif isinstance(reservation, ABCBrokerReservation):
            self.revisit_server(reservation=reservation)

    def revisit_client(self, *, reservation: ABCClientReservation):
        """
        Recovers a source reservation.

        @param reservation reservation to recover

        @throws Exception in case of error
        """
        if (reservation.get_state() == ReservationStates.Nascent and
            reservation.get_pending_state() == ReservationPendingStates.None_) or \
            (reservation.get_state() == ReservationStates.Ticketed and
             reservation.get_pending_state() == ReservationPendingStates.ExtendingTicket):
            self.calendar.add_pending(reservation=reservation)

    def revisit_server(self, *, reservation: ABCBrokerReservation):
        """
        Recovers a client reservation.

        @param reservation reservation to recover

        @throws Exception in case of error
        """
        if reservation.get_state() == ReservationStates.Ticketed and \
                (reservation.get_pending_state() == ReservationPendingStates.None_ or
                    reservation.get_pending_state() == ReservationPendingStates.Priming):
            source = reservation.get_source()
            if source is None:
                raise BrokerException(msg=f"source delegation is None for res# {reservation}")

            self.calendar.add_outlay(source=source,
                                     client=reservation,
                                     start=reservation.get_term().get_new_start_time(),
                                     end=reservation.get_term().get_end_time())

            self.calendar.add_closing(reservation=reservation,
                                      cycle=self.clock.cycle(when=reservation.get_term().get_end_time()))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from datetime import datetime
from typing import List


from fabric_mb.message_bus.messages.add_reservation_avro import AddReservationAvro
from fabric_mb.message_bus.messages.add_reservations_avro import AddReservationsAvro
from fabric_mb.message_bus.messages.claim_resources_avro import ClaimResourcesAvro
from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.demand_reservation_avro import DemandReservationAvro
from fabric_mb.message_bus.messages.extend_reservation_avro import ExtendReservationAvro
from fabric_mb.message_bus.messages.get_actors_request_avro import GetActorsRequestAvro
from fabric_mb.message_bus.messages.get_broker_query_model_request_avro import GetBrokerQueryModelRequestAvro
from fabric_mb.message_bus.messages.broker_query_model_avro import BrokerQueryModelAvro
from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
from fabric_mb.message_bus.messages.reclaim_resources_avro import ReclaimResourcesAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.reservation_predecessor_avro import ReservationPredecessorAvro
from fabric_mb.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
from fim.slivers.base_sliver import BaseSliver
from fim.user import GraphFormat

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.apis.abc_mgmt_broker_mixin import ABCMgmtBrokerMixin
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.manage.kafka.kafka_server_actor import KafkaServerActor
from fabric_cf.actor.core.util.id import ID


class KafkaBroker(KafkaServerActor, ABCMgmtBrokerMixin):
    def add_reservation(self, *, reservation: TicketReservationAvro) -> ID:
        request = AddReservationAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.message_id = str(ID())
        request.callback_topic = self.callback_topic
        request.reservation_obj = reservation
        status, response = self.send_request(request)

        if status.code == 0:
            return response.result_str

    def add_reservations(self, *, reservations: list) -> List[TicketReservationAvro]:
        request = AddReservationsAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.message_id = str(ID())
        request.callback_topic = self.callback_topic
        request.reservation_list = reservations
        status, response = self.send_request(request)

        if status.code == 0:
            return response.result

        return None

    def demand_reservation(self, *, reservation: ReservationMng) -> bool:
        request = DemandReservationAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.message_id = str(ID())
        request.callback_topic = self.callback_topic
        request.reservation_obj = reservation
        status, response = self.send_request(request)

        return status.code == 0

    def demand_reservation_rid(self, *, rid: ID) -> bool:
        request = DemandReservationAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.message_id = str(ID())
        request.callback_topic = self.callback_topic
        request.reservation_id = str(rid)
        status, response = self.send_request(request)

        return status.code == 0

    def get_brokers(self, *, broker: ID = None, id_token: str = None) -> List[ProxyAvro]:
        request = GetActorsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token, broker_id=broker)
        status, response = self.send_request(request)

        if status.code == 0:
            return response.proxies
        return None

    def get_broker_query_model(self, *, broker: ID, id_token: str, level: int,
                               graph_format: GraphFormat) -> BrokerQueryModelAvro:
        request = GetBrokerQueryModelRequestAvro()
        request.id_token = id_token
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.message_id = str(ID())
        request.callback_topic = self.callback_topic
        request.broker_id = str(broker)
        request.level = level
        request.graph_format = graph_format.value
        status, response = self.send_request(request)

        if status.code == 0:
            return response.model
        return None

    def extend_reservation(self, *, reservation: ID, new_end_time: datetime, sliver: BaseSliver,
                           dependencies: List[ReservationPredecessorAvro] = None) -> bool:
        request = ExtendReservationAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.message_id = str(ID())
        request.callback_topic = self.callback_topic
        request.rid = str(reservation)
        request.sliver = sliver

        status, response = self.send_request(request)

        return status.code == 0

    def claim_delegations(self, *, broker: ID, did: ID) -> DelegationAvro:

        request = ClaimResourcesAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.broker_id = str(broker)
        request.delegation_id = did
        request.message_id = str(ID())
        request.callback_topic = self.callback_topic

        status, response = self.send_request(request)

        if status.code == 0 and response.delegations is not None and len(response.delegations) > 0:
            return next(iter(response.delegations))

        return None

    def reclaim_delegations(self, *, broker: ID, did: ID) -> DelegationAvro:
        request = ReclaimResourcesAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.broker_id = str(broker)
        request.delegation_id = did
        request.message_id = str(ID())
        request.callback_topic = self.callback_topic

        status, response = self.send_request(request)

        if status.code == 0 and response.delegations is not None and len(response.delegations) > 0:
            return next(iter(response.delegations))

        return None

    def clone(self):
        return KafkaBroker(guid=self.management_id,
                           kafka_topic=self.kafka_topic,
                           auth=self.auth, logger=self.logger,
                           message_processor=self.message_processor,
                           producer=self.producer)

    def add_broker(self, *, broker: ProxyAvro) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from typing import List

from fabric_mb.message_bus.messages.lease_reservation_avro import LeaseReservationAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.slice_avro import SliceAvro
from fim.graph.abc_property_graph import ABCPropertyGraph

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates, JoinState
from fabric_cf.actor.core.kernel.slice_state_machine import SliceState
from fabric_cf.actor.core.time.actor_clock import ActorClock


class ResponseBuilder:
    PROP_SLICE_ID = "slice_id"
    PROP_NAME = "name"
    PROP_STATE = "state"
    PROP_PROJECT_ID = "project_id"
    PROP_PROJECT_NAME = "project_name"
    PROP_MODEL = "model"
    PROP_GRAPH_ID = "graph_id"
    PROP_LEASE_START_TIME = "lease_start_time"
    PROP_LEASE_END_TIME = "lease_end_time"

    PROP_SLIVER_ID = "sliver_id"
    PROP_PENDING_STATE = "pending_state"
    PROP_JOIN_STATE = "join_state"
    PROP_GRAPH_NODE_ID = "graph_node_id"
    PROP_SLIVER = "sliver"
    PROP_SLIVER_TYPE = "sliver_type"
    PROP_NOTICE = "notice"

    @staticmethod
    def get_reservation_summary(*, res_list: List[ReservationMng]) -> List[dict]:
        """
        Get Reservation summary
        :param res_list:
        :return:
        """
        reservations = []

        if res_list is not None:
            for reservation in res_list:
                res_dict = {ResponseBuilder.PROP_SLICE_ID: reservation.get_slice_id(),
                            ResponseBuilder.PROP_SLIVER_ID: reservation.get_reservation_id(),
                            ResponseBuilder.PROP_STATE: ReservationStates(reservation.get_state()).name}

                if reservation.get_pending_state() is not None:
                    res_dict[ResponseBuilder.PROP_PENDING_STATE] = \
                        ReservationPendingStates(reservation.get_pending_state()).name

                if isinstance(reservation, LeaseReservationAvro) and reservation.get_join_state() is not None:
                    res_dict[ResponseBuilder.PROP_JOIN_STATE] = JoinState(reservation.get_join_state()).name

                sliver = reservation.get_sliver()
                if sliver is not None:
                    res_dict[ResponseBuilder.PROP_GRAPH_NODE_ID] = sliver.node_id
                    res_dict[ResponseBuilder.PROP_SLIVER_TYPE] = type(sliver).__name__
                    res_dict[ResponseBuilder.PROP_SLIVER] = ABCPropertyGraph.sliver_to_dict(sliver)

                if reservation.get_start() is not None:
                    start_time = ActorClock.from_milliseconds(milli_seconds=reservation.get_start())
                    res_dict[ResponseBuilder.PROP_LEASE_START_TIME] = start_time.strftime(Constants.LEASE_TIME_FORMAT)

                if reservation.get_end() is not None:
                    end_time = ActorClock.from_milliseconds(milli_seconds=reservation.get_end())
                    res_dict[ResponseBuilder.PROP_LEASE_END_TIME] = end_time.strftime(Constants.LEASE_TIME_FORMAT)

                if reservation.get_notices() is not None:
                    res_dict[ResponseBuilder.PROP_NOTICE] = reservation.get_notices()

                reservations.append(res_dict)

        return reservations

    @staticmethod
    def get_slice_summary(*, slice_list: List[SliceAvro], slice_model: str = None) -> List[dict]:
        """
        Get slice summary
        :param slice_list:
        :param slice_model:
        :return:
        """
        slices = []

        if slice_list is None:
            return slices

        for s in slice_list:
            s_dict = {ResponseBuilder.PROP_SLICE_ID: s.get_slice_id(),
                      ResponseBuilder.PROP_NAME: s.get_slice_name(),
                      ResponseBuilder.PROP_GRAPH_ID: s.get_graph_id(),
                      ResponseBuilder.PROP_STATE: SliceState(s.get_state()).name,
                      }
            end_time = s.get_lease_end()
            if end_time is not None:
                s_dict[ResponseBuilder.PROP_LEASE_END_TIME] = end_time.strftime(Constants.LEASE_TIME_FORMAT)
            start_time = s.get_lease_start()
            if start_time is not None:
                s_dict[ResponseBuilder.PROP_LEASE_START_TIME] = start_time.strftime(Constants.LEASE_TIME_FORMAT)

            if s.get_project_id() is not None:
                s_dict[ResponseBuilder.PROP_PROJECT_ID] = s.get_project_id()

            if s.get_project_name() is not None:
                s_dict[ResponseBuilder.PROP_PROJECT_NAME] = s.get_project_name()

            if slice_model is not None:
                s_dict[ResponseBuilder.PROP_MODEL] = slice_model
            slices.append(s_dict)

        return slices

    @staticmethod
    def get_broker_query_model_summary(*, bqm: str):
        """
        Get BQM
        :param bqm:
        :return:
        """
        return {ResponseBuilder.PROP_MODEL: bqm}



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import enum
import logging
import queue
import threading
import time
import traceback

from fabric_cf.actor.core.apis.abc_actor_event import ABCActorEvent
from fabric_cf.actor.core.apis.abc_actor_runnable import ABCActorRunnable
from fabric_cf.actor.core.apis.abc_timer_task import ABCTimerTask
from fabric_cf.actor.core.util.iterable_queue import IterableQueue


class EventType(enum.Enum):
    TickEvent = enum.auto(),
    InterActorEvent = enum.auto(),
    SyncEvent = enum.auto()

    def __str__(self):
        return self.name


class EventProcessorException(Exception):
    pass


class ExecutionStatus:
    """
    Execution status
    """
    def __init__(self):
        self.done = False
        self.exception = None
        self.result = None
        self.lock = threading.Condition()

    def mark_done(self):
        """
        Mark as done
        """
        self.done = True


class TickEvent(ABCActorEvent):
    def __init__(self, *, actor, cycle: int):
        self.actor = actor
        self.cycle = cycle

    def __str__(self):
        return "{} {}".format(self.actor, self.cycle)

    def process(self):
        self.actor.actor_tick(cycle=self.cycle)


class ActorEvent(ABCActorEvent):
    def __init__(self, *, status: ExecutionStatus, runnable: ABCActorRunnable):
        self.status = status
        self.runnable = runnable

    def process(self):
        """
        Process an event
        """
        try:
            self.status.result = self.runnable.run()
        except Exception as e:
            traceback.print_exc()
            self.status.exception = e
        finally:
            with self.status.lock:
                self.status.done = True
                self.status.lock.notify_all()


class EventProcessor:
    def __init__(self, *, name, logger: logging.Logger = None):
        self.event_queue = queue.Queue()
        self.thread_lock = threading.Lock()
        self.condition = threading.Condition()
        self.thread = None
        self.shutdown = False
        self.name = name
        if logger is None:
            self.logger = logging.Logger(self.name)
        else:
            self.logger = logger

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['event_queue']
        del state['thread_lock']
        del state['condition']
        del state['thread']
        del state['shutdown']
        del state['logger']

    def __setstate__(self, state):
        self.event_queue = queue.Queue()
        self.thread_lock = threading.Lock()
        self.condition = threading.Condition()
        self.thread = None
        self.shutdown = False
        self.logger = None

    def set_logger(self, *, logger: logging.Logger):
        self.logger = logger

    def start(self):
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise EventProcessorException(f"Event Processor {self.name} has already been started")

            self.thread = threading.Thread(target=self.__run)
            self.thread.setName(self.name)
            self.thread.setDaemon(True)
            self.thread.start()
            self.logger.debug(f"Event Processor {self.name} has been started")
        finally:
            self.thread_lock.release()

    def stop(self):
        self.shutdown = True
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the Event Processor {self.name} thread is running. Interrupting it")
                try:
                    temp.join()
                except Exception as e:
                    self.logger.error(f"Could not join Event Processor {self.name} thread {e}")
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def enqueue(self, incoming):
        try:
            self.event_queue.put_nowait(incoming)
            with self.condition:
                self.condition.notify_all()
            self.logger.debug("Added event to event queue {}".format(incoming.__class__.__name__))
        except Exception as e:
            self.logger.error(f"Failed to queue event: {incoming.__class__.__name__} e: {e}")

    def execute_on_thread_async(self, *, runnable: ABCActorRunnable):
        """
        Execute an incoming action on actor thread
        @param runnable incoming action/operation
        """
        if self.__is_on_event_processor_thread():
            return runnable.run()
        else:
            status = ExecutionStatus()
            event = ActorEvent(status=status, runnable=runnable)

            self.enqueue(incoming=event)

    def execute_on_thread_sync(self, *, runnable: ABCActorRunnable):
        """
        Execute an incoming action on actor thread
        @param runnable incoming action/operation
        """
        if self.__is_on_event_processor_thread():
            return runnable.run()
        else:
            status = ExecutionStatus()
            event = ActorEvent(status=status, runnable=runnable)

            self.enqueue(incoming=event)

            with status.lock:
                while not status.done:
                    status.lock.wait()

            if status.exception is not None:
                raise status.exception

            return status.result

    def __dequeue(self, queue_obj: queue.Queue):
        events = []
        if not queue_obj.empty():
            try:
                for event in IterableQueue(source_queue=queue_obj):
                    events.append(event)
            except Exception as e:
                self.logger.error(f"Error while adding event/timer to queue! e: {e}")
                self.logger.error(traceback.format_exc())
        return events

    def __process_events(self, *, events: list):
        for event in events:
            try:
                begin = time.time()
                if isinstance(event, ABCTimerTask):
                    event.execute()
                else:
                    event.process()
                self.logger.info(f"Event {event.__class__.__name__} TIME: {time.time() - begin:.0f}")
            except Exception as e:
                self.logger.error(f"Error while processing event {type(event)}, {e}")
                self.logger.error(traceback.format_exc())

    def __run(self):
        while True:
            with self.condition:
                while not self.shutdown and self.event_queue.empty():
                    self.condition.wait()

            if self.shutdown:
                break

            if self.shutdown:
                self.logger.info(f"Event Processor {self.name} exiting")
                return

            events = self.__dequeue(self.event_queue)
            self.__process_events(events=events)

    def __is_on_event_processor_thread(self) -> bool:
        """
        Check if running on event_processor
        @return true if running on event_processor, false otherwise
        """
        result = False
        try:
            self.thread_lock.acquire()
            result = self.thread == threading.current_thread()
        finally:
            self.thread_lock.release()
        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import queue
import threading

from fim.graph.abc_property_graph import ABCPropertyGraph
from fim.graph.resources.neo4j_arm import Neo4jARMGraph

from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType
from fabric_cf.actor.core.apis.abc_authority import ABCAuthority
from fabric_cf.actor.core.apis.abc_authority_reservation import ABCAuthorityReservation
from fabric_cf.actor.core.apis.abc_client_callback_proxy import ABCClientCallbackProxy
from fabric_cf.actor.core.apis.abc_controller_callback_proxy import ABCControllerCallbackProxy
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import AuthorityException
from fabric_cf.actor.core.core.actor import ActorMixin
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.kernel.slice import SliceFactory
from fabric_cf.actor.core.manage.authority_management_object import AuthorityManagementObject
from fabric_cf.actor.core.manage.kafka.services.kafka_authority_service import KafkaAuthorityService
from fabric_cf.actor.core.proxies.kafka.services.authority_service import AuthorityService
from fabric_cf.actor.core.delegation.delegation_factory import DelegationFactory
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.util.client import Client
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.reservation_set import ReservationSet
from fabric_cf.actor.security.auth_token import AuthToken


class Authority(ActorMixin, ABCAuthority):
    """
    Authority is the base implementation for a site authority actor.
    """
    def __init__(self, *, identity: AuthToken = None, clock: ActorClock = None):
        super().__init__(auth=identity, clock=clock)
        self.type = ActorType.Authority
        # Initialization status.
        self.initialized = False
        # Reservations to redeem once the actor recovers.
        self.redeeming = ReservationSet()
        # Reservations to extendLease for once the actor recovers.
        self.extending_lease = ReservationSet()
        # Reservations to modifyLease for once the actor recovers
        self.modifying_lease = ReservationSet()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['recovered']
        del state['wrapper']
        del state['logger']
        del state['clock']
        del state['current_cycle']
        del state['first_tick']
        del state['stopped']
        del state['initialized']
        del state['closing']
        del state['message_service']

        del state['redeeming']
        del state['extending_lease']
        del state['modifying_lease']
        del state['event_processors']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.recovered = False
        self.wrapper = None
        self.logger = None
        self.clock = None
        self.current_cycle = -1
        self.first_tick = True
        self.stopped = False
        self.initialized = False
        self.closing = ReservationSet()
        self.message_service = None

        self.redeeming = ReservationSet()
        self.extending_lease = ReservationSet()
        self.modifying_lease = ReservationSet()
        self.event_processors = {}

    def register_client_slice(self, *, slice_obj: ABCSlice):
        self.wrapper.register_slice(slice_object=slice_obj)

    def claim_delegation(self, *, delegation: ABCDelegation, callback: ABCClientCallbackProxy, caller: AuthToken):
        slice_obj = delegation.get_slice_object()
        if slice_obj is not None:
            slice_obj.set_broker_client()

        self.wrapper.claim_delegation_request(delegation=delegation, caller=caller, callback=callback)

    def reclaim_delegation(self, *, delegation: ABCDelegation, callback: ABCClientCallbackProxy, caller: AuthToken):

        slice_obj = delegation.get_slice_object()
        if slice_obj is not None:
            slice_obj.set_broker_client()

        self.wrapper.reclaim_delegation_request(delegation=delegation, caller=caller, callback=callback)

    def close_by_caller(self, *, reservation: ABCReservationMixin, caller: AuthToken):
        if not self.is_recovered() or self.is_stopped():
            raise AuthorityException(Constants.INVALID_ACTOR_STATE)

        self.wrapper.close_request(reservation=reservation, caller=caller, compare_sequence_numbers=True)

    def close_expiring(self, *, cycle: int):
        """
        Closes the expiring reservations for the specified cycle.

        @param cycle
                   cycle number
        """
        expired = self.policy.get_closing(cycle=cycle)
        if expired is not None:
            # self.logger.info("Authority expiring for cycle {} = {}".format(cycle, expired))
            self.close_reservations(reservations=expired)

    def donate_delegation(self, *, delegation: ABCDelegation):
        self.policy.donate_delegation(delegation=delegation)

    def eject(self, *, resources: ResourceSet):
        self.policy.eject(resources=resources)

    def advertise(self, *, delegation: ABCPropertyGraph, delegation_name: str, client: AuthToken) -> str:
        slice_obj = SliceFactory.create(slice_id=ID(), name=client.get_name())
        slice_obj.set_owner(owner=client)
        slice_obj.set_broker_client()

        dlg_obj = DelegationFactory.create(did=delegation.get_graph_id(), slice_id=slice_obj.get_slice_id(),
                                           delegation_name=delegation_name)
        dlg_obj.set_slice_object(slice_object=slice_obj)
        dlg_obj.set_graph(graph=delegation)
        self.wrapper.advertise(delegation=dlg_obj, client=client)
        return dlg_obj.get_delegation_id()

    def extend_lease(self, *, reservation: ABCAuthorityReservation, caller: AuthToken = None,
                     callback: ABCControllerCallbackProxy = None,):
        if caller is None:
            if not self.recovered:
                self.extending_lease.add(reservation=reservation)
            else:
                self.wrapper.extend_lease_request(reservation=reservation, caller=reservation.get_client_auth_token(),
                                                  compare_sequence_numbers=False, callback=callback)
        else:
            if not self.is_recovered() or self.is_stopped():
                raise AuthorityException(Constants.INVALID_ACTOR_STATE)
            self.wrapper.extend_lease_request(reservation=reservation, caller=caller, compare_sequence_numbers=True,
                                              callback=callback)

    def modify_lease(self, *, reservation: ABCAuthorityReservation, caller: AuthToken,
                     callback: ABCControllerCallbackProxy = None,):
        if caller is None:
            if not self.recovered:
                self.modifying_lease.add(reservation=reservation)
            else:
                self.wrapper.modify_lease_request(reservation=reservation, caller=reservation.get_client_auth_token(),
                                                  compare_sequence_numbers=False, callback=callback)
        else:
            if not self.is_recovered() or self.stopped:
                raise AuthorityException(Constants.INVALID_ACTOR_STATE)
            self.wrapper.modify_lease_request(reservation=reservation, caller=caller, compare_sequence_numbers=True,
                                              callback=callback)

    def extend_ticket(self, *, reservation: ABCReservationMixin, callback: ABCClientCallbackProxy, caller: AuthToken):
        slice_obj = reservation.get_slice()
        if slice_obj is not None:
            slice_obj.set_broker_client()

        self.wrapper.extend_ticket_request(reservation=reservation, caller=caller, compare_sequence_numbers=True,
                                           callback=callback)

    def relinquish(self, *, reservation: ABCReservationMixin, caller: AuthToken):
        if not self.is_recovered() or self.stopped:
            raise AuthorityException(Constants.INVALID_ACTOR_STATE)
        self.wrapper.relinquish_request(reservation=reservation, caller=caller)

    def freed(self, *, resources: ResourceSet):
        self.policy.freed(resources=resources)

    def redeem(self, *, reservation: ABCReservationMixin, callback: ABCControllerCallbackProxy = None,
               caller: AuthToken = None):
        if callback is None and caller is None:
            if not self.recovered:
                self.redeeming.add(reservation=reservation)
            else:
                self.wrapper.redeem_request(reservation=reservation, caller=reservation.get_client_auth_token(),
                                            callback=reservation.get_callback(), compare_sequence_numbers=False)
        else:
            if not self.is_recovered() or self.is_stopped():
                raise AuthorityException(Constants.INVALID_ACTOR_STATE)

            if self.plugin.validate_incoming(reservation=reservation, auth=caller):
                self.wrapper.redeem_request(reservation=reservation, caller=caller, callback=callback,
                                            compare_sequence_numbers=True)
            else:
                self.logger.error("the redeem request is invalid")
        self.logger.debug("Completed processing Redeem Request")

    def ticket(self, *, reservation: ABCReservationMixin, callback: ABCClientCallbackProxy, caller: AuthToken):
        slice_obj = reservation.get_slice()
        if slice_obj is not None:
            slice_obj.set_broker_client()

        self.wrapper.ticket_request(reservation=reservation, caller=caller, callback=callback,
                                    compare_seq_numbers=True)

    def tick_handler(self):
        # close expired reservations
        self.close_expiring(cycle=self.current_cycle)
        # process all requests for the current cycle
        self.policy.assign(cycle=self.current_cycle)

    def register_client(self, *, client: Client):
        db = self.plugin.get_database()

        try:
            db.get_client(guid=client.get_guid())
        except Exception as e:
            self.logger.debug("Client does not exist e:{}".format(e))

        db.add_client(client=client)

    def unregister_client(self, *, guid: ID):
        db = self.plugin.get_database()
        db.remove_client(guid=guid)

    def get_client(self, *, guid: ID) -> Client:
        db = self.plugin.get_database()

        return db.get_client(guid=guid)

    def redeem_reservations(self, *, rset: ReservationSet):
        """
        Redeem all reservations:
        @param rset: reservation set
        """
        for reservation in rset.values():
            try:
                if isinstance(reservation, ABCAuthorityReservation):
                    self.redeem(reservation=reservation)
                else:
                    self.logger.warning("Reservation # {} cannot be redeemed".format(reservation.get_reservation_id()))
            except Exception as e:
                self.logger.error("Could not redeem for # {} {}".format(reservation.get_reservation_id(), e))

    def extend_lease_reservations(self, *, rset: ReservationSet):
        """
        Extend all reservations:
        @param rset: reservation set
        """
        for reservation in rset.values():
            try:
                self.extend_lease(reservation=reservation)
            except Exception as e:
                self.logger.error("Could not redeem for # {} {}".format(reservation.get_reservation_id(), e))

    def issue_delayed(self):
        super().issue_delayed()
        self.redeem_reservations(rset=self.redeeming)
        self.redeeming.clear()
        self.extend_lease_reservations(rset=self.extending_lease)
        self.extending_lease.clear()

    @staticmethod
    def get_management_object_class() -> str:
        return AuthorityManagementObject.__name__

    @staticmethod
    def get_management_object_module() -> str:
        return AuthorityManagementObject.__module__

    @staticmethod
    def get_kafka_service_class() -> str:
        return AuthorityService.__name__

    @staticmethod
    def get_kafka_service_module() -> str:
        return AuthorityService.__module__

    @staticmethod
    def get_mgmt_kafka_service_class() -> str:
        return KafkaAuthorityService.__name__

    @staticmethod
    def get_mgmt_kafka_service_module() -> str:
        return KafkaAuthorityService.__module__

    def set_aggregate_resource_model(self, aggregate_resource_model: Neo4jARMGraph):
        """
        Set aggregate resource model
        :param aggregate_resource_model: resource model
        :return:
        """
        self.policy.set_aggregate_resource_model(aggregate_resource_model=aggregate_resource_model)

    def load_model(self, *, graph_id: str):
        self.policy.set_aggregate_resource_model_graph_id(graph_id=graph_id)
        self.policy.load_aggregate_resource_model()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
import traceback
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_mgmt_client_actor import ABCMgmtClientActor
from fabric_cf.actor.core.apis.abc_mgmt_server_actor import ABCMgmtServerActor
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.actor_identity import ActorIdentity
from fabric_cf.actor.core.manage.messages.client_mng import ClientMng
from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_mgmt_actor import ABCMgmtActor
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin


class RemoteActorCacheException(Exception):
    """
    Exception raised by RemoteActorCache
    """


class RemoteActorCache:
    """
    Maintains Remote Actors to which Actor is connected
    """
    actor_name = "NAME"
    actor_guid = "GUID"
    actor_location = "LOCATION"
    actor_protocol = "PROTOCOL"
    actor_type = "TYPE"

    def __init__(self):
        self.cache = {}
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.lock = threading.Lock()
        self.local_actor_guids = set()

    def known_guids(self) -> set:
        """
        Returns list of know GUIDs
        """
        try:
            self.lock.acquire()
            result = set()
            for k in self.cache.keys():
                result.add(k)
            return result
        finally:
            self.lock.release()

    def check_to_remove_entry(self, *, guid: ID):
        """
        Check if actor entry can be removed
        @param guid actor guid
        """
        if guid is None:
            return

        try:
            self.lock.acquire()
            if guid in self.cache:
                value = self.cache[guid]
                if self.actor_location not in value:
                    self.cache.pop(guid)
        finally:
            self.lock.release()

    def add_cache_entry(self, *, guid: ID, entry: dict):
        """
        Add actor entry to Cache
        @param guid actor guid
        @param entry actor entry
        """
        if guid is None or entry is None:
            return

        try:
            self.lock.acquire()
            self.non_mt_cache_merge(guid=guid, entry=entry)
            self.local_actor_guids.add(guid)
        finally:
            self.lock.release()

    def add_partial_cache_entry(self, *, guid: ID, entry: dict):
        """
        Add partial actor entry to Cache
        @param guid actor guid
        @param entry actor entry
        """
        if guid is None or entry is None:
            return
        try:
            self.lock.acquire()
            self.non_mt_cache_merge(guid=guid, entry=entry)
        finally:
            self.lock.release()

    def non_mt_cache_merge(self, *, guid: ID, entry: dict):
        """
        Merge to an existing entry if present; remove otherwise
        @param guid actor guid
        @param entry actor entry
        """
        if guid not in self.cache:
            self.logger.debug("Inserting new entry for {}".format(guid))
            self.cache[guid] = entry
            return
        current = self.cache[guid]
        for k, v in entry.items():
            current[k] = v

        self.cache[guid] = current

    def get_cache_entry_copy(self, *, guid: ID) -> dict:
        """
        Get a copy of cacher entry
        @param guid actor guid
        """
        try:
            self.lock.acquire()
            if guid in self.cache:
                ret_val = self.cache[guid]
                return ret_val
        finally:
            self.lock.release()

    def check_peer(self, *, from_mgmt_actor: ABCMgmtActor, from_guid: ID, to_mgmt_actor: ABCMgmtActor, to_guid: ID):
        """
        Check if a peer is already connected
        @param from_mgmt_actor from actor
        @param from_guid guid of from actor
        @param to_mgmt_actor to actor
        @param to_guid guid of to actor
        """
        self.logger.debug("from_mgmt_actor={} from_guid={} to_mgmt_actor={} to_guid={}".format(type(from_mgmt_actor),
                                                                                               from_guid,
                                                                                               type(to_mgmt_actor),
                                                                                               to_guid))
        try:
            # For Broker/AM
            if to_mgmt_actor is not None and isinstance(to_mgmt_actor, ABCMgmtServerActor):
                clients = to_mgmt_actor.get_clients(guid=from_guid)
                if clients is not None:
                    self.logger.debug("Edge between {} and {} exists (client)".format(from_guid, to_guid))
                    return True

            # For Orchestrator/Broker
            elif from_mgmt_actor is not None and isinstance(from_mgmt_actor, ABCMgmtClientActor):
                brokers = from_mgmt_actor.get_brokers(broker=to_guid)
                if brokers is not None:
                    self.logger.debug("Edge between {} and {} exists (broker)".format(from_guid, to_guid))
                    return True
        except Exception as e:
            raise RemoteActorCacheException("Unable to cast actor {} or {} e={}".format(from_guid, to_guid, e))

        self.logger.debug("Edge between {} and {} does not exist".format(from_guid, to_guid))
        return False

    def establish_peer_private(self, *, from_mgmt_actor: ABCMgmtActor, from_guid: ID, to_mgmt_actor: ABCMgmtActor,
                               to_guid: ID) -> ClientMng:
        """
        Establish connection i.e. create either proxies or clients between peer
        @param from_mgmt_actor from actor
        @param from_guid guid of from actor
        @param to_mgmt_actor to actor
        @param to_guid guid of to actor
        """
        self.logger.debug("establish_peer_private IN")
        client = None
        from_map = self.get_cache_entry_copy(guid=from_guid)
        to_map = self.get_cache_entry_copy(guid=to_guid)

        if from_map is None:
            raise RemoteActorCacheException("Actor {} does not have a registry cache entry".format(from_guid))

        if to_map is None:
            raise RemoteActorCacheException("Actor {} does not have a registry cache entry".format(to_guid))

        if from_mgmt_actor is not None:
            self.logger.debug("From actor {} is local".format(from_mgmt_actor.get_name()))

            protocol = Constants.PROTOCOL_LOCAL
            kafka_topic = None

            if self.actor_location in to_map:
                if self.actor_protocol not in to_map:
                    raise RemoteActorCacheException("Actor {} does not specify communications protocol (local/kafka)".
                                                    format(to_map[self.actor_name]))

                protocol = to_map.get(self.actor_protocol, None)
                kafka_topic = to_map[self.actor_location]
                self.logger.debug("Added To actor location (non-local) {}".format(to_map[self.actor_location]))

            identity = ActorIdentity(name=to_map[self.actor_name], guid=to_guid)

            if kafka_topic is not None and isinstance(from_mgmt_actor, ABCMgmtClientActor):
                self.logger.debug("Kafka Topic is available, registering broker proxy")
                proxy = ProxyAvro()
                proxy.set_protocol(protocol)
                proxy.set_guid(str(identity.get_guid()))
                proxy.set_name(identity.get_name())
                proxy.set_type(to_map[self.actor_type])
                proxy.set_kafka_topic(kafka_topic)

                try:
                    if not from_mgmt_actor.add_broker(broker=proxy):
                        raise RemoteActorCacheException("Could not register broker {}".
                                                        format(from_mgmt_actor.get_last_error()))
                except Exception as e:
                    self.logger.error(e)
                    self.logger.error(traceback.format_exc())
            else:
                self.logger.debug("Not adding broker to actor at this time because the remote actor actor "
                                  "kafka topic is not available")

            if to_mgmt_actor is not None and isinstance(to_mgmt_actor, ABCMgmtServerActor):
                self.logger.debug("Creating a client for local to actor")
                client = ClientMng()
                client.set_name(name=from_mgmt_actor.get_name())
                client.set_guid(guid=str(from_mgmt_actor.get_guid()))
                try:
                    to_mgmt_actor.register_client(client=client, kafka_topic=kafka_topic)
                except Exception as e:
                    raise RemoteActorCacheException("Could not register actor: {} as a client of actor: {} e= {}".
                                                    format(client.get_name(), to_mgmt_actor.get_name(), e))
        else:
            # fromActor is remote: toActor must be local
            # no-need to create any proxies
            # we only need to register clients
            if to_mgmt_actor is None:
                raise RemoteActorCacheException("Both peer endpoints are non local actors: {} {}".format(
                    from_map[self.actor_name], to_map[self.actor_name]))

            if self.actor_guid not in from_map:
                raise RemoteActorCacheException("Missing guid for remote actor: {}".format(from_map[self.actor_name]))

            self.logger.debug("From actor was remote, to actor {} is local".format(to_mgmt_actor.get_name()))
            if self.actor_location in from_map and isinstance(to_mgmt_actor, ABCMgmtServerActor):
                kafka_topic = from_map[self.actor_location]
                self.logger.debug("From actor has kafka topic")
                self.logger.debug("Creating client for from actor {}".format(from_map[self.actor_name]))
                client = ClientMng()
                client.set_name(name=from_map[self.actor_name])
                client.set_guid(guid=str(from_map[self.actor_guid]))
                try:
                    to_mgmt_actor.register_client(client=client, kafka_topic=kafka_topic)
                except Exception as e:
                    raise RemoteActorCacheException(
                        "Could not register actor: {} as a client of actor: {} e= {}".format(
                            client.get_name(), to_mgmt_actor.get_name(), e))
            else:
                self.logger.debug("Not adding client to actor at this time - remote actor topic not available")
        self.logger.debug("establish_peer_private OUT {}".format(client))
        return client

    def establish_peer(self, *, from_guid: ID, from_mgmt_actor: ABCMgmtActor, to_guid: ID,
                       to_mgmt_actor: ABCMgmtActor) -> ClientMng:
        """
        Check if peer exists in cache and if not Establish connection i.e. create either proxies or clients between peer
        @param from_mgmt_actor from actor
        @param from_guid guid of from actor
        @param to_mgmt_actor to actor
        @param to_guid guid of to actor
        """
        self.logger.debug("establish_peer IN")
        client = None
        if from_guid is None or to_guid is None:
            self.logger.error("Cannot establish peer when either guid is not known")
            raise RemoteActorCacheException("Cannot establish peer when either guid is not known")
        try:
            if not self.check_peer(from_mgmt_actor=from_mgmt_actor, from_guid=from_guid,
                                   to_mgmt_actor=to_mgmt_actor, to_guid=to_guid):

                client = self.establish_peer_private(from_mgmt_actor=from_mgmt_actor, from_guid=from_guid,
                                                     to_mgmt_actor=to_mgmt_actor, to_guid=to_guid)

                self.check_to_remove_entry(guid=from_guid)
                self.check_to_remove_entry(guid=to_guid)

                self.logger.debug("Peer established from {} to {}".format(from_guid, to_guid))

        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Peer could not be established from {} to {} e:={}".format(from_guid, to_guid, e))
        self.logger.debug("establish_peer OUT {}".format(client))
        return client

    def register_with_registry(self, *, actor: ABCActorMixin):
        """
        Register an actor with Registry
        @param actor actor
        """
        try:
            act_name = actor.get_name()
            act_type = actor.get_type()
            act_guid = actor.get_guid()

            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            kafka_topic = GlobalsSingleton.get().get_config().get_actor_config().get_kafka_topic()

            entry = {self.actor_name: act_name,
                     self.actor_guid: act_guid,
                     self.actor_type: act_type.name,
                     self.actor_protocol: Constants.PROTOCOL_KAFKA,
                     self.actor_location: kafka_topic}

            self.add_cache_entry(guid=act_guid, entry=entry)
            # TODO start liveness thread
        except Exception as e:
            self.logger.debug("Could not register actor {} with local registry e: {}".format(actor.get_name(), e))


class RemoteActorCacheSingleton:
    """
    Remote Actor Cache Singleton
    """
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise RemoteActorCacheException("Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = RemoteActorCache()
        return self.__instance

    get = classmethod(get)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from fabric_mb.message_bus.messages.slice_avro import SliceAvro

from fabric_cf.actor.core.apis.abc_query_response_handler import ABCQueryResponseHandler
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.apis.abc_mgmt_container import ABCMgmtContainer
from fabric_cf.actor.core.manage.local.local_container import LocalContainer
from fabric_cf.actor.core.proxies.kafka.translate import Translate
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.rpc_exception import RPCException, RPCError
from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.reservation_mng import ReservationMng

    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.apis.abc_actor_proxy import ABCActorProxy
    from fabric_cf.actor.security.auth_token import AuthToken

from fabric_cf.actor.core.manage.converter import Converter


class MyQueryResponseHandler(ABCQueryResponseHandler):
    def __init__(self):
        self.result = None
        self.error = None
        self.condition = threading.Condition()

    def handle(self, *, status: RPCException, result: dict):
        with self.condition:
            self.error = status
            self.result = result
            self.condition.notify()


class ManagementUtils:
    @staticmethod
    def update_slice(*, slice_obj: ABCSlice, slice_mng: SliceAvro) -> ABCSlice:
        slice_obj.set_graph_id(graph_id=slice_mng.get_graph_id())
        slice_obj.set_lease_end(lease_end=slice_mng.get_lease_end())
        slice_obj.set_config_properties(value=slice_mng.get_config_properties())
        return slice_obj

    @staticmethod
    def update_reservation(*, res_obj: ABCReservationMixin, rsv_mng: ReservationMng) -> ABCReservationMixin:
        if isinstance(res_obj, ABCClientReservation):
            res_obj.set_renewable(renewable=rsv_mng.is_renewable())

        return Converter.absorb_res_properties(rsv_mng=rsv_mng, res_obj=res_obj)

    @staticmethod
    def query(*, actor: ABCActorMixin, actor_proxy: ABCActorProxy, query: dict):
        handler = MyQueryResponseHandler()

        actor.query(query=query, actor_proxy=actor_proxy, handler=handler)

        with handler.condition:
            try:
                handler.condition.wait()
                # TODO InterruptedException
            except Exception as e:
                raise RPCException(message=str(e), error=RPCError.LocalError)

        if handler.error is not None:
            raise handler.error

        return handler.result

    @staticmethod
    def get_local_actor():
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        local_actor = GlobalsSingleton.get().get_container().get_actor()
        local_mgmt_container = ManagementUtils.get_local_container(caller=local_actor.get_identity())
        return local_mgmt_container.get_actor(guid=local_actor.get_guid())

    @staticmethod
    def get_local_container(*, caller: AuthToken):
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        manager = GlobalsSingleton.get().get_container().get_management_object_manager().get_management_object(
            key=ID(uid=Constants.CONTAINER_MANAGMENT_OBJECT_ID))
        proxy = LocalContainer(manager=manager, auth=caller)
        return proxy

    @staticmethod
    def connect(*, caller: AuthToken) -> ABCMgmtContainer():
        return ManagementUtils.get_local_container(caller=caller)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from fabric_mb.message_bus.messages.abc_message_avro import AbcMessageAvro

from fabric_cf.actor.core.apis.abc_reservation_mixin import ReservationCategory
from fabric_cf.actor.core.kernel.slice import SliceTypes
from fabric_cf.actor.core.manage.kafka.services.kafka_client_actor_service import KafkaClientActorService
from fabric_cf.actor.core.manage.kafka.services.kafka_server_actor_service import KafkaServerActorService


class KafkaBrokerService(KafkaClientActorService, KafkaServerActorService):
    def process(self, *, message: AbcMessageAvro):
        callback_topic = message.get_callback_topic()
        result = None

        self.logger.debug("Processing message: {}".format(message.get_message_name()))
        result = self.authorize_request(id_token=message.get_id_token(), message_name=message.get_message_name())

        # If authorization failed, return the result
        if result is None:
            if message.get_message_name() == AbcMessageAvro.claim_resources:
                result = self.claim(request=message)

            elif message.get_message_name() == AbcMessageAvro.reclaim_resources:
                result = self.reclaim(request=message)

            elif message.get_message_name() == AbcMessageAvro.add_reservation:
                result = self.add_reservation(request=message)

            elif message.get_message_name() == AbcMessageAvro.add_reservations:
                result = self.add_reservations(request=message)

            elif message.get_message_name() == AbcMessageAvro.demand_reservation:
                result = self.demand_reservation(request=message)

            elif message.get_message_name() == AbcMessageAvro.get_actors_request:
                result = self.get_brokers(request=message)

            elif message.get_message_name() == AbcMessageAvro.get_broker_query_model_request:
                result = self.get_broker_query_model(request=message)

            elif message.get_message_name() == AbcMessageAvro.extend_reservation:
                result = self.extend_reservation(request=message)

            elif message.get_message_name() == AbcMessageAvro.get_reservations_request and \
                    message.get_type() is not None and \
                    message.get_type() == ReservationCategory.Broker.name:
                result = self.get_reservations_by_category(request=message, category=ReservationCategory.Broker)

            elif message.get_message_name() == AbcMessageAvro.get_slices_request and \
                    message.get_type() is not None and \
                    message.get_type() == SliceTypes.InventorySlice.name:
                result = self.get_slices_by_slice_type(request=message, slice_type=SliceTypes.InventorySlice)

            elif message.get_message_name() == AbcMessageAvro.get_reservations_request and \
                    message.get_type() is not None and \
                    message.get_type() == ReservationCategory.Inventory.name:
                result = self.get_reservations_by_category(request=message, category=ReservationCategory.Inventory)

            elif message.get_message_name() == AbcMessageAvro.get_slices_request and \
                    message.get_type() is not None and \
                    message.get_type() == SliceTypes.ClientSlice.name:
                result = self.get_slices_by_slice_type(request=message, slice_type=SliceTypes.ClientSlice)

            elif message.get_message_name() == AbcMessageAvro.add_slice and message.slice_obj is not None and \
                    (message.slice_obj.is_client_slice() or message.slice_obj.is_broker_client_slice()):
                result = self.add_client_slice(request=message)

            elif message.get_message_name() == AbcMessageAvro.remove_delegation:
                result = self.remove_delegation(request=message)

            elif message.get_message_name() == AbcMessageAvro.close_delegations:
                result = self.close_delegations(request=message)

            else:
                super().process(message=message)
                return

        if callback_topic is None:
            self.logger.debug("No callback specified, ignoring the message")

        if result is None:
            self.logger.error("No response generated {}".format(result))
            return

        if self.producer.produce(topic=callback_topic, record=result):
            self.logger.debug("Successfully send back response: {}".format(result.to_dict()))
        else:
            self.logger.debug("Failed to send back response: {}".format(result.to_dict()))



from fabric_cf.orchestrator.swagger_server.models.resources import Resources  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response import resources_controller as rc


def portalresources_get(graph_format):  # noqa: E501
    """Retrieve a listing and description of available resources for portal

    Retrieve a listing and description of available resources for portal # noqa: E501

    :param graph_format: graph format
    :type graph_format: str

    :rtype: Resources
    """
    return rc.portalresources_get(graph_format)


def resources_get(level, force_refresh):  # noqa: E501
    """Retrieve a listing and description of available resources. By default, a cached available resource information is returned. User can force to request the current available resources.

    Retrieve a listing and description of available resources. By default, a cached available resource information is returned. User can force to request the current available resources. # noqa: E501

    :param level: Level of details
    :type level: int
    :param force_refresh: Force to retrieve current available resource information.
    :type force_refresh: bool

    :rtype: Resources
    """
    return rc.resources_get(level, force_refresh)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_reservation_resources import ABCReservationResources
from fabric_cf.actor.core.apis.abc_reservation_status import ABCReservationStatus


if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.util.reservation_state import ReservationState
    from fabric_cf.actor.core.util.resource_type import ResourceType
    from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
    from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
    from fabric_cf.actor.core.kernel.failed_rpc import FailedRPC
    from fabric_cf.actor.core.kernel.request_types import RequestTypes
    from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
    from fabric_cf.actor.core.util.update_data import UpdateData


class ReservationCategory(Enum):
    """
    Enumeration for Reservation Category
    """
    # Unspecified reservation category.
    All = 0
    # Client-side reservations.
    Client = 1
    # Broker-side reservations.
    Broker = 2
    # Site authority-side reservations.
    Authority = 3
    # Used only for Get Operations
    Inventory = 4


class ABCReservationMixin(ABCReservationResources, ABCReservationStatus):
    """
    IReservation defines the the core API for a reservation. Most of the methods described in the interface allow the
    programmer to inspect the state of the reservation, access some of its core objects,
    and wait for the occurrence of a particular event.
    """
    @abstractmethod
    def clear_dirty(self):
        """
        Marks that the reservation has no uncommitted updates or state transitions.
        """

    @abstractmethod
    def get_actor(self) -> ABCActorMixin:
        """
        Returns the actor in control of the reservation.

        Returns:
            the actor in control of the reservation.
        """

    @abstractmethod
    def get_category(self) -> ReservationCategory:
        """
        Returns the reservation category.

        Returns:
            the reservation category.
        """

    @abstractmethod
    def get_pending_state(self) -> ReservationPendingStates:
        """
        Returns the current pending reservation state.

        Returns:
            current pending reservation state.
        """

    @abstractmethod
    def get_pending_state_name(self) -> str:
        """
        Returns the current pending reservation state name

        Returns:
            current pending reservation state name
        """

    @abstractmethod
    def get_reservation_id(self) -> ID:
        """
        Returns the reservation id.

        Returns:
            the reservation id.
        """

    @abstractmethod
    def get_reservation_state(self) -> ReservationState:
        """
        Returns the current composite reservation state.

        Returns:
            current composite reservation state.
        """

    @abstractmethod
    def get_slice(self) -> ABCSlice:
        """
        Returns the slice the reservation belongs to.

        Returns:
            slice the reservation belongs to.
        """

    @abstractmethod
    def get_slice_id(self) -> ID:
        """
        Returns the slice GUID.

        Returns:
            slice guid
        """

    @abstractmethod
    def get_slice_name(self) -> str:
        """
        Returns the slice name.

        Returns:
            slice name
        """

    @abstractmethod
    def get_state(self) -> ReservationStates:
        """
        Returns the current reservation state.

        Returns:
            current reservation state.
        """

    @abstractmethod
    def get_state_name(self) -> str:
        """
        Returns the current reservation state name.

        Returns:
            current reservation state name.
        """

    @abstractmethod
    def get_type(self) -> ResourceType:
        """
        Returns the resource type allocated to the reservation. If no resources have yet been allocated to the
        reservation, this method will return None.

        Returns:
            resource type allocated to the reservation. None if no resources have been allocated to the reservation.
        """

    @abstractmethod
    def has_uncommitted_transition(self) -> bool:
        """
        Checks if the reservation has uncommitted state transitions.

        Returns:
            true if the reservation has an uncommitted transition
        """

    @abstractmethod
    def is_dirty(self) -> bool:
        """
        Checks if the reservation has uncommitted updates.

        Returns:
            true if the reservation has an uncommitted updates
        """

    @abstractmethod
    def is_pending_recover(self) -> bool:
        """
        Checks if a recovery operation is in progress for the reservation

        Returns:
            true if a recovery operation for the reservation is in progress
        """

    @abstractmethod
    def set_dirty(self):
        """
        Marks the reservation as containing uncommitted updates.
        """

    @abstractmethod
    def set_pending_recover(self, *, pending_recover: bool):
        """
        Indicates if a recovery operation for the reservation is going to be in progress.

        Args:
            pending_recover: true, a recovery operation is in progress, false - no recovery operation is in progress.
        """

    @abstractmethod
    def transition(self, *, prefix: str, state: ReservationStates, pending: ReservationPendingStates):
        """
        Transitions this reservation into a new state.

        Args:
            prefix: prefix
            state: the new state
            pending: if reservation is pending
        """

    @abstractmethod
    def get_notices(self) -> str:
        """
        Returns the error message associated with this reservation.
        @return error message associated with this reservation
        """

    @abstractmethod
    def get_graph_node_id(self) -> str:
        """
        Returns the graph node id which represents the sliver for this reservation in the graph model
        @return graph node id
        """

    @abstractmethod
    def get_error_message(self) -> str:
        """
        Return Error Message
        """

    @abstractmethod
    def set_slice(self, *, slice_object: ABCSlice):
        """
        Sets the slice the reservation belongs to.

        @param slice_object slice the reservation belongs to
        """

    @abstractmethod
    def can_redeem(self) -> bool:
        """
        Checks if the reservation can be redeemed at the current time.

        @return true if the reservation's current state allows it to be redeemed
        """

    @abstractmethod
    def can_renew(self) -> bool:
        """
        Checks if this reservation can be renewed at the current time.

        @return true if the reservation's current state allows it to be renewed
        """

    @abstractmethod
    def close(self):
        """
        Closes the reservation. Locked with the kernel lock.
        """

    @abstractmethod
    def extend_lease(self):
        """
        Extends the reservation.

        @throws Exception in case of error
        """

    @abstractmethod
    def modify_lease(self):
        """
        Modifies the reservation.

        @throws Exception in case of error
        """

    @abstractmethod
    def extend_ticket(self, *, actor: ABCActorMixin):
        """
        Extends the ticket.

        @param actor actor

        @throws Exception in case of error
        """

    @abstractmethod
    def get_kernel_slice(self):
        """
        Returns the kernel slice.

        @return kernel slice
        """

    @abstractmethod
    def handle_duplicate_request(self, *, operation: RequestTypes):
        """
        Handles a duplicate request.

        @param operation operation type code
        @throws Exception in case of error
        """

    @abstractmethod
    def prepare(self, *, callback: ABCCallbackProxy, logger):
        """
        Prepares for a ticket request on a new reservation object.

        @param callback callback object
        @param logger for diagnostic logging
        @throws Exception in case of error
        """

    @abstractmethod
    def prepare_probe(self):
        """
        Prepares a reservation probe.

        @throws Exception in case of error
        """

    @abstractmethod
    def probe_pending(self):
        """
        Probe a reservation with a pending request. On server, if the
        operation completed, handle it and generate an update. If no pending
        request completed then do nothing.

        @throws Exception in case of error
        """

    @abstractmethod
    def reserve(self, *, policy: ABCPolicy):
        """
        Reserve resources: ticket() initiate or request, or redeem()
        request. New reservation.

        @param policy the mapper for the reservation

        @throws Exception in case of error
        """

    @abstractmethod
    def service_claim(self):
        """
        Finishes processing claim.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_reclaim(self):
        """
        Finishes processing reclaim.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_close(self):
        """
        Finishes processing close.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_extend_lease(self):
        """
        Finishes processing extend lease.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_modify_lease(self):
        """
        Finishes processing modify lease.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_extend_ticket(self):
        """
        Finishes processing extend ticket.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_probe(self):
        """
        Finishes processing probe.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_reserve(self):
        """
        Finishes processing reserve.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_update_lease(self):
        """
        Finishes processing update lease.
        @throws Exception in case of error
        """

    @abstractmethod
    def service_update_ticket(self):
        """
        Finishes processing update ticket.
        @throws Exception in case of error
        """

    @abstractmethod
    def set_actor(self, *, actor: ABCActorMixin):
        """
        Sets the actor in control of the reservation.

        @param actor actor in control of the reservation
        """

    @abstractmethod
    def set_logger(self, *, logger):
        """
        Attaches the logger to use for the reservation.

        @param logger logger object
        """

    @abstractmethod
    def set_service_pending(self, *, code: int):
        """
        Indicates there is a pending operation on the reservation.

        @param code operation code
        """

    @abstractmethod
    def update_lease(self, *, incoming: ABCReservationMixin, update_data: UpdateData):
        """
        Handles an incoming lease update.

        @param incoming incoming lease update
        @param update_data update data

        @throws Exception thrown if lease update is for an un-ticketed
                reservation
        @throws Exception thrown if lease update is for closed reservation
        """

    @abstractmethod
    def update_ticket(self, *, incoming: ABCReservationMixin, update_data: UpdateData):
        """
        Handles an incoming ticket update.

        @param incoming incoming ticket update
        @param update_data update data

        @throws Exception in case of error
        """

    @abstractmethod
    def validate_incoming(self):
        """
        Validates a reservation as it arrives at an actor.
        @throws Exception in case of error
        """

    @abstractmethod
    def validate_outgoing(self):
        """
        Validates a reservation as it is about to leave an actor.

        @throws Exception in case of error
        """

    @abstractmethod
    def handle_failed_rpc(self, *, failed: FailedRPC):
        """
        Processes a failed RPC request.
        @param failed failed
        """

    @abstractmethod
    def lock(self):
        """
        Lock the reservation
        """

    @abstractmethod
    def unlock(self):
        """
        Unlock the reservation
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime, timezone

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import BrokerException
from fabric_cf.actor.core.kernel.broker_query_model_publisher import BrokerQueryModelPublisher


class BrokerKernel:
    """
    Class responsible for starting Broker Periodic; also holds Management Actor
    """

    def __init__(self):
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.broker = GlobalsSingleton.get().get_container().get_actor()
        self.producer = GlobalsSingleton.get().get_simple_kafka_producer()
        self.kafka_topic = GlobalsSingleton.get().get_config().get_global_config().get_bqm_config().get(
            Constants.KAFKA_TOPIC, None)
        self.publish_interval = GlobalsSingleton.get().get_config().get_global_config().get_bqm_config().get(
            Constants.PUBLISH_INTERVAL, None)
        self.last_query_time = None

    def do_periodic(self):
        """
        Periodically publish BQM to a Kafka Topic to be consumed by Portal
        """
        if self.kafka_topic is not None and self.publish_interval is not None and self.producer is not None:
            current_time = datetime.now(timezone.utc)
            if self.last_query_time is None or (current_time - self.last_query_time).seconds > self.publish_interval:
                bqm = BrokerQueryModelPublisher(broker=self.broker, logger=self.logger,
                                                kafka_topic=self.kafka_topic, producer=self.producer)
                bqm.execute()
                self.last_query_time = datetime.now(timezone.utc)


class BrokerKernelSingleton:
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise BrokerException(msg="Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = BrokerKernel()
        return self.__instance

    get = classmethod(get)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import time
import traceback
from typing import Tuple

from fim.slivers.attached_components import ComponentType
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver, ServiceType

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.plugins.handlers.config_token import ConfigToken
from fabric_cf.actor.core.util.utils import sliver_to_str
from fabric_cf.actor.handlers.handler_base import HandlerBase


class NoOpHandler(HandlerBase):
    def __process_node_sliver_modify(self, *, existing: NodeSliver, new: NodeSliver):
        # Validate both existing and new sliver
        self.__process_node_sliver(sliver=existing)
        self.__process_node_sliver(sliver=new)
        #diff = sliver_diff(sliver1=existing, sliver2=new)
        diff = existing.diff(other_sliver=new)
        for x in diff.added.components:
            # invoke PCI attach
            self.get_logger().info(f"PCI attach for component {x}")

        for x in diff.removed.components:
            # invoke PCI detach
            self.get_logger().info(f"PCI detach for component {x}")

    def __process_node_sliver(self, *, sliver: NodeSliver):
        assert (sliver.label_allocations is not None)
        assert (sliver.label_allocations.instance_parent is not None)
        assert (sliver.get_capacity_hints() is not None)
        assert (sliver.get_capacity_hints().instance_type is not None)
        assert (sliver.get_image_ref() is not None)
        assert (sliver.get_name() is not None)
        assert (sliver.get_type() is not None)
        if sliver.attached_components_info is not None:
            for component in sliver.attached_components_info.devices.values():
                assert (component.label_allocations is not None)
                if component.get_type() != ComponentType.Storage:
                    assert (component.label_allocations.bdf is not None)
                else:
                    assert (component.label_allocations.local_name is not None)
                    component.label_allocations.device_name = '/dev/vdb'
        sliver.label_allocations.instance = 'instance_001'
        sliver.management_ip = '1.2.3.4'

        nic_types = [ComponentType.SharedNIC, ComponentType.SmartNIC]
        if sliver.attached_components_info is not None:
            for c in sliver.attached_components_info.devices.values():
                if c.get_type() not in nic_types:
                    continue
                self.get_logger().debug(f"c: {c}")
                if c.network_service_info is not None and c.network_service_info.network_services is not None:
                    for ns in c.network_service_info.network_services.values():
                        self.get_logger().debug(f"ns: {ns}")
                        if ns.interface_info is not None and ns.interface_info.interfaces is not None:
                            for i in ns.interface_info.interfaces.values():
                                self.get_logger().debug(f"ifs: {i}")

    def __process_ns_sliver(self, *, sliver: NetworkServiceSliver):
        assert (sliver.get_type() is not None)
        assert (sliver.interface_info is not None)
        for interface in sliver.interface_info.interfaces.values():
            assert (interface.labels is not None)
            assert (interface.labels.device_name is not None)
            assert (interface.labels.local_name is not None)
            assert (interface.capacities is not None)

            if sliver.get_type() == ServiceType.L2PTP:
                assert (interface.labels.vlan is not None)

            if sliver.get_type() in Constants.L3_FABNETv4_SERVICES:
                assert (sliver.get_gateway() is not None)
                assert (sliver.get_gateway().lab.ipv4_subnet is not None)
                assert (sliver.get_gateway().lab.ipv4 is not None)
                assert (interface.labels.vlan is not None)

            if sliver.get_type() in Constants.L3_FABNETv6_SERVICES:
                assert (sliver.get_gateway() is not None)
                assert (sliver.get_gateway().lab.ipv6_subnet is not None)
                assert (sliver.get_gateway().lab.ipv6 is not None)
                assert (interface.labels.vlan is not None)

            if sliver.get_type() == ServiceType.L3VPN:
                assert (sliver.labels.asn is not None)
                assert (interface.labels.vlan is not None)
                assert (interface.labels.local_name is not None)
                assert (interface.labels.device_name is not None)
                assert (interface.labels.ipv4_subnet is not None)

    def create(self, unit: ConfigToken) -> Tuple[dict, ConfigToken]:
        result = None
        try:
            self.get_logger().info(f"Create invoked for unit: {unit}")
            sliver = unit.get_sliver()
            self.get_logger().info(f"Creating sliver: {sliver_to_str(sliver=sliver)}")

            time.sleep(10)

            if isinstance(sliver, NodeSliver):
                self.__process_node_sliver(sliver=sliver)

            elif isinstance(sliver, NetworkServiceSliver):
                self.__process_ns_sliver(sliver=sliver)

            result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_CREATE,
                      Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_OK,
                      Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}
        except Exception as e:
            result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_CREATE,
                      Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_EXCEPTION,
                      Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}
            self.get_logger().error(e)
            self.get_logger().error(traceback.format_exc())
        finally:

            self.get_logger().info(f"Create completed")
        return result, unit

    def delete(self, unit: ConfigToken) -> Tuple[dict, ConfigToken]:
        result = None
        try:
            self.get_logger().info(f"Delete invoked for unit: {unit}")
            result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_DELETE,
                      Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_OK,
                      Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}
        except Exception as e:
            result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_DELETE,
                      Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_EXCEPTION,
                      Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}
            self.get_logger().error(e)
            self.get_logger().error(traceback.format_exc())
        finally:

            self.get_logger().info(f"Delete completed")
        return result, unit

    def modify(self, unit: ConfigToken) -> Tuple[dict, ConfigToken]:
        result = None
        try:
            self.get_logger().info(f"Modify invoked for unit: {unit} type{unit.get_modified()}")
            sliver = unit.get_modified()
            self.get_logger().info(f"Modifying sliver: {sliver_to_str(sliver=sliver)}")

            if isinstance(sliver, NodeSliver):
                self.__process_node_sliver_modify(existing=unit.get_sliver(), new=sliver)

            elif isinstance(sliver, NetworkServiceSliver):
                self.__process_ns_sliver(sliver=sliver)

            result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_MODIFY,
                      Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_OK,
                      Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}
        except Exception as e:
            self.get_logger().error(e)
            self.get_logger().error(traceback.format_exc())
            result = {Constants.PROPERTY_TARGET_NAME: Constants.TARGET_MODIFY,
                      Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_EXCEPTION,
                      Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}
        finally:
            self.get_logger().info(f"Modify completed")
        return result, unit

    def clean_restart(self):
        pass
